from typing import Any, List, Optional
from uuid import UUID

from koda import Just, Maybe, nothing

from koda_validate import Predicate, PredicateAsync, Processor
from koda_validate._internal import _ToTupleStandardValidator
from koda_validate.coerce import Coercer, coercer


@coercer(str, UUID)
def coerce_uuid(val: Any) -> Maybe[UUID]:
    if type(val) is UUID:
        return Just(val)

    elif type(val) is str:
        try:
            return Just(UUID(val))
        except ValueError:
            pass

    return nothing


class UUIDValidator(_ToTupleStandardValidator[UUID]):
    _TYPE = UUID

    def __init__(
        self,
        *predicates: Predicate[UUID],
        predicates_async: Optional[List[PredicateAsync[UUID]]] = None,
        preprocessors: Optional[List[Processor[UUID]]] = None,
        coerce: Optional[Coercer[UUID]] = coerce_uuid,
    ) -> None:
        super().__init__(
            *predicates,
            predicates_async=predicates_async,
            preprocessors=preprocessors,
            coerce=coerce,
        )



from datetime import date, datetime
from typing import Any, List, Optional

from koda import Just, Maybe, nothing

from koda_validate import Predicate, PredicateAsync, Processor
from koda_validate._internal import _ToTupleStandardValidator
from koda_validate.coerce import Coercer, coercer


@coercer(str, date)
def coerce_date(val: Any) -> Maybe[date]:
    if type(val) is date:
        return Just(val)
    else:
        try:
            return Just(date.fromisoformat(val))
        except (ValueError, TypeError):
            return nothing


class DateValidator(_ToTupleStandardValidator[date]):
    _TYPE = date

    def __init__(
        self,
        *predicates: Predicate[date],
        predicates_async: Optional[List[PredicateAsync[date]]] = None,
        preprocessors: Optional[List[Processor[date]]] = None,
        coerce: Optional[Coercer[date]] = coerce_date,
    ) -> None:
        super().__init__(
            *predicates,
            predicates_async=predicates_async,
            preprocessors=preprocessors,
            coerce=coerce,
        )


@coercer(str, datetime)
def coerce_datetime(val: Any) -> Maybe[datetime]:
    if type(val) is datetime:
        return Just(val)
    else:
        try:
            return Just(datetime.fromisoformat(val))
        except (ValueError, TypeError):
            return nothing


class DatetimeValidator(_ToTupleStandardValidator[datetime]):
    _TYPE = datetime

    def __init__(
        self,
        *predicates: Predicate[datetime],
        predicates_async: Optional[List[PredicateAsync[datetime]]] = None,
        preprocessors: Optional[List[Processor[datetime]]] = None,
        coerce: Optional[Coercer[datetime]] = coerce_datetime,
    ) -> None:
        super().__init__(
            *predicates,
            predicates_async=predicates_async,
            preprocessors=preprocessors,
            coerce=coerce,
        )



import importlib.metadata

__version__ = importlib.metadata.version("koda_validate")

__all__ = (
    # base.py
    "Validator",
    "Predicate",
    "PredicateAsync",
    "Processor",
    # boolean.py
    "BoolValidator",
    # bytes.py
    # coerce.py
    "Coercer",
    "coercer",
    "BytesValidator",
    # dataclasses.py
    "DataclassValidator",
    # decimal.py
    "DecimalValidator",
    # dictionary.py
    "KeyNotRequired",
    "MapValidator",
    "is_dict_validator",
    "IsDictValidator",
    "MinKeys",
    "MaxKeys",
    "RecordValidator",
    "DictValidatorAny",
    # errors.py
    "CoercionErr",
    "ContainerErr",
    "ExtraKeysErr",
    "ErrType",
    "IndexErrs",
    "KeyErrs",
    "KeyValErrs",
    "MapErr",
    "MissingKeyErr",
    "missing_key_err",
    "PredicateErrs",
    "SetErrs",
    "TypeErr",
    "UnionErrs",
    "ValidationErrBase",
    # float.py
    "FloatValidator",
    # generic.py
    "Lazy",
    "Choices",
    "Min",
    "Max",
    "MinItems",
    "MaxItems",
    "ExactItemCount",
    "unique_items",
    "UniqueItems",
    "MultipleOf",
    "EqualsValidator",
    "EqualTo",
    "always_valid",
    "AlwaysValid",
    "MinLength",
    "MaxLength",
    "ExactLength",
    "StartsWith",
    "EndsWith",
    "strip",
    "not_blank",
    "NotBlank",
    "upper_case",
    "UpperCase",
    "lower_case",
    "LowerCase",
    "CacheValidatorBase",
    # integer.py
    "IntValidator",
    # list.py
    "ListValidator",
    # namedtuple.py
    "NamedTupleValidator",
    # none.py
    "OptionalValidator",
    "NoneValidator",
    "none_validator",
    # set.py
    "SetValidator",
    # string.py
    "StringValidator",
    "RegexPredicate",
    "EmailPredicate",
    # time.py
    "DateValidator",
    "DatetimeValidator",
    # tuple.py
    "NTupleValidator",
    "UniformTupleValidator",
    # typeddict.py
    "TypedDictValidator",
    # uuid.py
    "UUIDValidator",
    # union.py
    "UnionValidator",
    # valid.py
    "Valid",
    "Invalid",
    "ValidationResult",
)

from koda_validate.base import (
    CacheValidatorBase,
    Predicate,
    PredicateAsync,
    Processor,
    Validator,
)
from koda_validate.boolean import BoolValidator
from koda_validate.bytes import BytesValidator
from koda_validate.coerce import Coercer, coercer
from koda_validate.dataclasses import DataclassValidator
from koda_validate.decimal import DecimalValidator
from koda_validate.dictionary import (
    DictValidatorAny,
    IsDictValidator,
    KeyNotRequired,
    MapValidator,
    MaxKeys,
    MinKeys,
    RecordValidator,
    is_dict_validator,
)
from koda_validate.errors import (
    CoercionErr,
    ContainerErr,
    ErrType,
    ExtraKeysErr,
    IndexErrs,
    KeyErrs,
    KeyValErrs,
    MapErr,
    MissingKeyErr,
    PredicateErrs,
    SetErrs,
    TypeErr,
    UnionErrs,
    ValidationErrBase,
    missing_key_err,
)
from koda_validate.float import FloatValidator
from koda_validate.generic import (
    AlwaysValid,
    Choices,
    EndsWith,
    EqualsValidator,
    EqualTo,
    ExactItemCount,
    ExactLength,
    Lazy,
    LowerCase,
    Max,
    MaxItems,
    MaxLength,
    Min,
    MinItems,
    MinLength,
    MultipleOf,
    NotBlank,
    StartsWith,
    UniqueItems,
    UpperCase,
    always_valid,
    lower_case,
    not_blank,
    strip,
    unique_items,
    upper_case,
)
from koda_validate.integer import IntValidator
from koda_validate.list import ListValidator
from koda_validate.namedtuple import NamedTupleValidator
from koda_validate.none import NoneValidator, OptionalValidator, none_validator
from koda_validate.set import SetValidator
from koda_validate.string import EmailPredicate, RegexPredicate, StringValidator
from koda_validate.time import DatetimeValidator, DateValidator
from koda_validate.tuple import NTupleValidator, UniformTupleValidator
from koda_validate.typeddict import TypedDictValidator
from koda_validate.union import UnionValidator
from koda_validate.uuid import UUIDValidator
from koda_validate.valid import Invalid, Valid, ValidationResult



import functools
import inspect
from dataclasses import is_dataclass
from datetime import date, datetime
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
    overload,
)
from uuid import UUID

from _decimal import Decimal

from koda_validate import DataclassValidator, NamedTupleValidator
from koda_validate._internal import _is_typed_dict_cls
from koda_validate.base import Validator
from koda_validate.dataclasses import dataclass_no_coerce
from koda_validate.errors import (
    CoercionErr,
    ContainerErr,
    ExtraKeysErr,
    IndexErrs,
    KeyErrs,
    MapErr,
    MissingKeyErr,
    PredicateErrs,
    SetErrs,
    TypeErr,
    UnionErrs,
)
from koda_validate.generic import always_valid
from koda_validate.namedtuple import namedtuple_no_coerce
from koda_validate.time import DatetimeValidator, DateValidator
from koda_validate.tuple import NTupleValidator, UniformTupleValidator
from koda_validate.typeddict import TypedDictValidator
from koda_validate.typehints import (
    annotation_is_naked_tuple,
    annotation_is_namedtuple,
    get_typehint_validator_base,
)
from koda_validate.uuid import UUIDValidator
from koda_validate.valid import Invalid

_BaseDecoratedFunc = Callable[..., Any]
_DecoratedFunc = TypeVar("_DecoratedFunc", bound=_BaseDecoratedFunc)

# for overrides - just to avoid parameter name conflicts
ReturnOverrideKey = Tuple[Literal["return_key"]]
RETURN_OVERRIDE_KEY: ReturnOverrideKey = ("return_key",)

OverridesDict = Dict[Union[str, ReturnOverrideKey], Validator[Any]]


def resolve_signature_typehint_default(annotation: Any) -> Validator[Any]:
    if annotation is Decimal:
        from koda_validate.decimal import DecimalValidator

        return DecimalValidator(coerce=None)
    elif _is_typed_dict_cls(annotation):
        return TypedDictValidator(annotation)
    elif annotation is UUID:
        return UUIDValidator(coerce=None)
    elif annotation is date:
        return DateValidator(coerce=None)
    elif annotation is datetime:
        return DatetimeValidator(coerce=None)
    elif annotation_is_naked_tuple(annotation):
        return UniformTupleValidator(always_valid, coerce=None)
    elif is_dataclass(annotation):
        return DataclassValidator(annotation, coerce=dataclass_no_coerce(annotation))
    elif annotation_is_namedtuple(annotation):
        return NamedTupleValidator(annotation, coerce=namedtuple_no_coerce(annotation))
    else:
        origin, args = get_origin(annotation), get_args(annotation)
        if annotation_is_naked_tuple(origin):
            if len(args) == 2 and args[1] is Ellipsis:
                return UniformTupleValidator(
                    resolve_signature_typehint_default(args[0]), coerce=None
                )
            else:
                return NTupleValidator.untyped(
                    fields=tuple(resolve_signature_typehint_default(a) for a in args),
                    coerce=None,
                )

    return get_typehint_validator_base(resolve_signature_typehint_default, annotation)


def _get_validator(
    overrides: OverridesDict,
    typehint_resolver: Callable[[Any], Validator[Any]],
    param_name: str,
    annotation: Any,
) -> Validator[Any]:
    return (
        overrides[param_name]
        if param_name in overrides
        else typehint_resolver(annotation)
    )


def _wrap_fn(
    func: _DecoratedFunc,
    ignore_args: Set[str],
    ignore_return: bool,
    typehint_resolver: Callable[[Any], Validator[Any]],
    overrides: OverridesDict,
) -> _DecoratedFunc:
    sig = inspect.signature(func)
    # This value is Optional because we need to keep track of all
    # the argument names. For arguments that either a) don't have
    # an annotation, or b) are being ignore, we keep the value as None
    # If we simply didn't store the keys of arguments we're ignoring, we
    # wouldn't be able to tell the difference between a **kwargs key and a
    # defined argument name
    schema: Dict[str, Optional[Validator[Any]]] = {}

    kwargs_validator: Optional[Validator[Any]] = None
    var_args_key_and_validator: Optional[Tuple[str, Validator[Any]]] = None
    positional_args_names: Set[str] = set()
    positional_validators: List[Optional[Tuple[str, Validator[Any]]]] = []
    _get_validator_partial = functools.partial(
        _get_validator, overrides, typehint_resolver
    )

    # create arg schema
    for key, param in sig.parameters.items():
        annotation = param.annotation
        if param.kind == param.POSITIONAL_ONLY:
            positional_args_names.add(key)
            if annotation == param.empty or key in ignore_args:
                positional_validators.append(None)
            else:
                positional_validators.append(
                    (key, _get_validator_partial(key, annotation))
                )

            schema[key] = None
        elif param.kind == param.POSITIONAL_OR_KEYWORD:
            positional_args_names.add(key)
            if annotation == param.empty or key in ignore_args:
                positional_validators.append(None)
                schema[key] = None
            else:
                validator = _get_validator_partial(key, annotation)
                positional_validators.append((key, validator))
                schema[key] = validator
        elif param.kind == param.VAR_POSITIONAL:
            if annotation != param.empty and key not in ignore_args:
                var_args_key_and_validator = key, _get_validator_partial(key, annotation)
        elif param.kind == param.KEYWORD_ONLY:
            if annotation != param.empty and key not in ignore_args:
                schema[key] = _get_validator_partial(key, annotation)
            else:
                schema[key] = None
        elif param.kind == param.VAR_KEYWORD:
            if annotation != param.empty and key not in ignore_args:
                kwargs_validator = _get_validator_partial(key, annotation)

    if not ignore_return and sig.return_annotation != sig.empty:
        return_validator: Optional[Validator[Any]] = _get_validator_partial(
            RETURN_OVERRIDE_KEY, sig.return_annotation
        )
    else:
        return_validator = None

    # we can allow keys only sent in **kwargs to be ignored as well
    ignored_extra_kwargs = ignore_args.difference(schema)

    if inspect.iscoroutinefunction(func) or inspect.iscoroutinefunction(
        getattr(func, "__call__", None)
    ):

        async def inner_async(*args: Any, **kwargs: Any) -> Any:
            errs: Dict[str, Invalid] = {}
            var_args_errs: List[Tuple[Any, Invalid]] = []
            # in case the values get mutated during validation
            ok_args: List[Any] = list(args)
            ok_kw_args: Dict[str, Any] = kwargs.copy()
            for i, arg in enumerate(args):
                if len(positional_validators) >= i + 1:
                    arg_details = positional_validators[i]
                    if arg_details is None:
                        # not an annotated argument
                        continue
                    else:
                        key, arg_validator = arg_details
                        if not (
                            result := await arg_validator.validate_async(arg)
                        ).is_valid:
                            errs[key] = result
                        else:
                            ok_args[i] = result.val

                else:
                    if var_args_key_and_validator:
                        var_args_key, var_args_validator = var_args_key_and_validator
                        if not (
                            var_args_result := await var_args_validator.validate_async(
                                arg
                            )
                        ).is_valid:
                            var_args_errs.append((arg, var_args_result))
                        else:
                            ok_args[i] = var_args_result.val

            if var_args_errs and var_args_key_and_validator:
                errs[var_args_key_and_validator[0]] = Invalid(
                    IndexErrs({i: err_ for i, (_, err_) in enumerate(var_args_errs)}),
                    tuple(a for a, _ in var_args_errs),
                    var_args_key_and_validator[1],
                )

            for kw_key, kw_val in kwargs.items():
                if kw_key in schema:
                    if schema_validator := schema[kw_key]:
                        if not (
                            kw_result := await schema_validator.validate_async(kw_val)
                        ).is_valid:
                            errs[kw_key] = kw_result
                        else:
                            ok_kw_args[kw_key] = kw_result.val
                elif kwargs_validator and kw_key not in ignored_extra_kwargs:
                    if not (
                        kwargs_result := await kwargs_validator.validate_async(kw_val)
                    ).is_valid:
                        errs[kw_key] = kwargs_result
                    else:
                        ok_kw_args[kw_key] = kwargs_result.val

            if errs:
                raise InvalidArgsError(errs)
            elif return_validator:
                result = await func(*ok_args, **ok_kw_args)
                if (ret_result := await return_validator.validate_async(result)).is_valid:
                    return result
                else:
                    raise InvalidReturnError(ret_result)
            else:
                return await func(*args, **kwargs)

        return cast(_DecoratedFunc, inner_async)

    else:

        @functools.wraps(func)
        def inner(*args: Any, **kwargs: Any) -> Any:
            errs: Dict[str, Invalid] = {}
            var_args_errs: List[Tuple[Any, Invalid]] = []
            # in case the values get mutated during validation
            ok_args: List[Any] = list(args)
            ok_kw_args: Dict[str, Any] = kwargs.copy()
            for i, arg in enumerate(args):
                if len(positional_validators) >= i + 1:
                    arg_details = positional_validators[i]
                    if arg_details is None:
                        # not an annotated argument
                        continue
                    else:
                        key, arg_validator = arg_details
                        if not (result := arg_validator(arg)).is_valid:
                            errs[key] = result
                        else:
                            ok_args[i] = result.val

                else:
                    if var_args_key_and_validator:
                        var_args_key, var_args_validator = var_args_key_and_validator
                        if not (var_args_result := var_args_validator(arg)).is_valid:
                            var_args_errs.append((arg, var_args_result))
                        else:
                            ok_args[i] = var_args_result.val

            if var_args_errs and var_args_key_and_validator:
                errs[var_args_key_and_validator[0]] = Invalid(
                    IndexErrs({i: err_ for i, (_, err_) in enumerate(var_args_errs)}),
                    tuple(a for a, _ in var_args_errs),
                    var_args_key_and_validator[1],
                )

            for kw_key, kw_val in kwargs.items():
                if kw_key in schema:
                    if schema_validator := schema[kw_key]:
                        if not (kw_result := schema_validator(kw_val)).is_valid:
                            errs[kw_key] = kw_result
                        else:
                            ok_kw_args[kw_key] = kw_result.val
                elif kwargs_validator and kw_key not in ignored_extra_kwargs:
                    if (kwargs_result := kwargs_validator(kw_val)).is_valid:
                        ok_kw_args[kw_key] = kwargs_result.val
                    else:
                        errs[kw_key] = kwargs_result

            if errs:
                raise InvalidArgsError(errs)
            elif return_validator:
                result = func(*ok_args, **ok_kw_args)
                if (ret_result := return_validator(result)).is_valid:
                    return result
                else:
                    raise InvalidReturnError(ret_result)
            else:
                return func(*ok_args, **kwargs)

        return cast(_DecoratedFunc, inner)


@overload
def validate_signature(
    func: _DecoratedFunc,
    *,
    ignore_args: Optional[Set[str]] = None,
    ignore_return: bool = False,
    typehint_resolver: Callable[
        [Any], Validator[Any]
    ] = resolve_signature_typehint_default,  # noqa: E501
    overrides: Optional[OverridesDict] = None,
) -> _DecoratedFunc:
    ...


@overload
def validate_signature(
    func: None = None,
    *,
    ignore_args: Optional[Set[str]] = None,
    ignore_return: bool = False,
    typehint_resolver: Callable[
        [Any], Validator[Any]
    ] = resolve_signature_typehint_default,  # noqa: E501
    overrides: Optional[OverridesDict] = None,
) -> Callable[[_DecoratedFunc], _DecoratedFunc]:
    ...


def validate_signature(
    func: Optional[_DecoratedFunc] = None,
    *,
    ignore_return: bool = False,
    ignore_args: Optional[Set[str]] = None,
    typehint_resolver: Callable[
        [Any], Validator[Any]
    ] = resolve_signature_typehint_default,  # noqa: E501
    overrides: Optional[OverridesDict] = None,
) -> Union[_DecoratedFunc, Callable[[_DecoratedFunc], _DecoratedFunc]]:
    r"""
    Validates a function's arguments and / or return value adhere to the respective
    typehints and / or any custom-specified Validation.Because we want to preserve the
    type signature of the function when it's wrapped, we raise exceptions to represent
    failure (as opposed to returning ``ValidationResult[ReturnType]``, which would change
    the function's type signature).

    :param func: the function being validated
    :param ignore_return: whether or not to ignore the return annotation
    :param ignore_args: any arguments that should be ignored
    :param typehint_resolver: the function responsible for resolving type annotations
        to :class:`koda_validate.Validator`\s
    :param overrides: explicit `Validator`s for arguments that takes priority over
        `typehint_resolver` and `Annotated` types
    :return: the decorated function
    """
    _wrap_fn_partial = functools.partial(
        _wrap_fn,
        ignore_return=ignore_return,
        ignore_args=ignore_args or set(),
        typehint_resolver=typehint_resolver,
        overrides=overrides or {},
    )

    if func is None:

        def inner(func_inner: _DecoratedFunc) -> _DecoratedFunc:
            # there may be a good way to replace this cast with ParamSpec in the future
            return cast(_DecoratedFunc, _wrap_fn_partial(func_inner))

        return inner
    else:
        # there may be a good way to replace this cast with ParamSpec in the future
        return cast(_DecoratedFunc, _wrap_fn_partial(func))


def _trunc_str(s: str, max_chars: int) -> str:
    ellip = "..."
    ellip_len = len(ellip)
    if max_chars < ellip_len:
        raise AssertionError(f"max_chars must be greater than or equal to {ellip_len}")

    return (s[: (max_chars - ellip_len)] + ellip) if len(s) > max_chars else s


def _get_arg_fail_message(invalid: Invalid, indent: str = "", prefix: str = "") -> str:
    err_type = invalid.err_type
    next_indent = f"    {indent}"
    ret = indent + prefix
    if isinstance(err_type, TypeErr):
        ret += f"expected {err_type.expected_type}"
    elif isinstance(err_type, PredicateErrs):
        ret += f"{err_type.__class__.__name__}\n"
        ret += "\n".join([f"{next_indent}{repr(p)}" for p in err_type.predicates])
    elif isinstance(err_type, CoercionErr):
        ret += (
            f"expected any of "
            f"{sorted(err_type.compatible_types, key=lambda t: repr(t))} to coerce "
            f"to {repr(err_type.dest_type)}"
        )
    elif isinstance(err_type, ContainerErr):
        return _get_arg_fail_message(err_type.child, prefix)
    elif isinstance(err_type, MissingKeyErr):
        ret += "key missing"
    elif isinstance(err_type, UnionErrs):
        variant_errors = [
            _get_arg_fail_message(variant, next_indent, prefix=f"variant {i + 1}: ")
            for i, variant in enumerate(err_type.variants)
        ]
        ret += "\n".join([err_type.__class__.__name__] + variant_errors)
    elif isinstance(err_type, KeyErrs):
        ret += f"{err_type.__class__.__name__}\n"
        ret += "\n".join(
            [
                _get_arg_fail_message(inv, next_indent, prefix=f"{repr(k)}: ")
                for k, inv in err_type.keys.items()
            ]
        )
    elif isinstance(err_type, ExtraKeysErr):
        ret += f"{err_type.__class__.__name__}\n"
        ret += (
            f"{next_indent}only expected keys {sorted(err_type.expected_keys, key=repr)}"
        )
    elif isinstance(err_type, IndexErrs):
        ret += f"{err_type.__class__.__name__}\n"
        ret += "\n".join(
            [
                _get_arg_fail_message(inv, next_indent, prefix=f"{idx}: ")
                for idx, inv in err_type.indexes.items()
            ]
        )
    elif isinstance(err_type, MapErr):
        ret += "MapErr"
        for key, key_val_errs in err_type.keys.items():
            if key_val_errs.key:
                next_ = _get_arg_fail_message(
                    key_val_errs.key, next_indent, prefix=f"{repr(key)} (key): "
                )
                ret += f"\n{next_}"
            if key_val_errs.val:
                next_ = _get_arg_fail_message(
                    key_val_errs.val, next_indent, prefix=f"{repr(key)} (val): "
                )
                ret += f"\n{next_}"
    elif isinstance(err_type, SetErrs):
        ret += f"{err_type.__class__.__name__}\n"
        ret += "\n".join(
            [
                f"{_get_arg_fail_message(e, next_indent)} :: {_trunc_str(repr(e.value), 30)}"  # noqa: E501
                # noqa: E501
                for e in err_type.item_errs
            ]
        )

    return ret


def _get_args_fail_msg(errs: Dict[str, Invalid]) -> str:
    messages = [
        f"{k}={_trunc_str(repr(v.value), 60)}\n{_get_arg_fail_message(v, '    ')}"
        for k, v in errs.items()
    ]
    return "\n".join(messages)


_INVALID_ARGS_MESSAGE_HEADER = "\nInvalid Argument Values\n-----------------------\n"
_INVALID_RETURN_MESSAGE_HEADER = "\nInvalid Return Value\n--------------------\n"


class InvalidArgsError(Exception):
    """
    Represents the validation failure of one or more arguments.
    """

    def __init__(self, errs: Dict[str, Invalid]) -> None:
        super().__init__(_INVALID_ARGS_MESSAGE_HEADER + _get_args_fail_msg(errs))
        self.errs = errs


class InvalidReturnError(Exception):
    """
    Represents a return value that has failed validation.
    """

    def __init__(self, err: Invalid):
        super().__init__(_INVALID_RETURN_MESSAGE_HEADER + _get_arg_fail_message(err))
        self.err = err



from typing import Any, Callable, Dict, List, Optional, Tuple, Union, overload

from koda import Just, Maybe, nothing

from koda_validate._generics import T1, T2, T3, T4, T5, T6, T7, T8, A
from koda_validate._internal import (
    _async_predicates_warning,
    _repr_helper,
    _ResultTuple,
    _ToTupleValidator,
    _wrap_async_validator,
    _wrap_sync_validator,
)
from koda_validate.base import Predicate, PredicateAsync, Validator
from koda_validate.coerce import Coercer, coercer
from koda_validate.errors import CoercionErr, ErrType, IndexErrs, PredicateErrs, TypeErr
from koda_validate.generic import ExactItemCount
from koda_validate.valid import Invalid


@coercer(list, tuple)
def tuple_or_list_to_tuple(val: Any) -> Maybe[Tuple[Any, ...]]:
    if (val_type := type(val)) is tuple:
        return Just(val)
    elif val_type is list:
        return Just(tuple(val))
    else:
        return nothing


class NTupleValidator(_ToTupleValidator[A]):
    __match_args__ = ("fields", "coerce")

    def __init__(
        self,
        *,
        fields: Tuple[Validator[Any], ...],
        validate_object: Optional[Callable[[A], Optional[ErrType]]] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> None:
        """
        You probably don't want to be using __init__ directly. For type-hinting reasons,
        it's recommended to use either `.typed(...)` or `.untyped(...)`

        :param fields: the validators that correspond with the respective indexes in the
            tuple
        :param validate_object: if all the slots of the tuple are valid, this can be used
            to validate among all the fields.
        :param coerce: control if an how values are coerced for this validator
        """

        self.fields = fields
        self.validate_object = validate_object
        self.coerce = coerce
        self._len_predicate = ExactItemCount(len(fields))
        self._wrapped_fields_sync = [_wrap_sync_validator(v) for v in fields]
        self._wrapped_fields_async = [_wrap_async_validator(v) for v in fields]

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[Validator[T1]],
        validate_object: Optional[Callable[[Tuple[T1]], Optional[ErrType]]] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[Validator[T1], Validator[T2]],
        validate_object: Optional[Callable[[Tuple[T1, T2]], Optional[ErrType]]] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[Validator[T1], Validator[T2], Validator[T3]],
        validate_object: Optional[
            Callable[[Tuple[T1, T2, T3]], Optional[ErrType]]
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2, T3]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[Validator[T1], Validator[T2], Validator[T3], Validator[T4]],
        validate_object: Optional[
            Callable[[Tuple[T1, T2, T3, T4]], Optional[ErrType]]
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2, T3, T4]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[
            Validator[T1],
            Validator[T2],
            Validator[T3],
            Validator[T4],
            Validator[T5],
        ],
        validate_object: Optional[
            Callable[[Tuple[T1, T2, T3, T4, T5]], Optional[ErrType]]
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2, T3, T4, T5]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[
            Validator[T1],
            Validator[T2],
            Validator[T3],
            Validator[T4],
            Validator[T5],
            Validator[T6],
        ],
        validate_object: Optional[
            Callable[[Tuple[T1, T2, T3, T4, T5, T6]], Optional[ErrType]]
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2, T3, T4, T5, T6]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[
            Validator[T1],
            Validator[T2],
            Validator[T3],
            Validator[T4],
            Validator[T5],
            Validator[T6],
            Validator[T7],
        ],
        validate_object: Optional[
            Callable[[Tuple[T1, T2, T3, T4, T5, T6, T7]], Optional[ErrType]]
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2, T3, T4, T5, T6, T7]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        *,
        fields: Tuple[
            Validator[T1],
            Validator[T2],
            Validator[T3],
            Validator[T4],
            Validator[T5],
            Validator[T6],
            Validator[T7],
            Validator[T8],
        ],
        validate_object: Optional[
            Callable[[Tuple[T1, T2, T3, T4, T5, T6, T7, T8]], Optional[ErrType]]
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[T1, T2, T3, T4, T5, T6, T7, T8]]":
        ...  # pragma: no cover

    @staticmethod
    def typed(
        *,
        fields: Union[
            Tuple[Validator[T1]],
            Tuple[Validator[T1], Validator[T2]],
            Tuple[Validator[T1], Validator[T2], Validator[T3]],
            Tuple[
                Validator[T1],
                Validator[T2],
                Validator[T3],
                Validator[T4],
            ],
            Tuple[
                Validator[T1], Validator[T2], Validator[T3], Validator[T4], Validator[T5]
            ],
            Tuple[
                Validator[T1],
                Validator[T2],
                Validator[T3],
                Validator[T4],
                Validator[T5],
                Validator[T6],
            ],
            Tuple[
                Validator[T1],
                Validator[T2],
                Validator[T3],
                Validator[T4],
                Validator[T5],
                Validator[T6],
                Validator[T7],
            ],
            Tuple[
                Validator[T1],
                Validator[T2],
                Validator[T3],
                Validator[T4],
                Validator[T5],
                Validator[T6],
                Validator[T7],
                Validator[T8],
            ],
        ],
        validate_object: Union[
            Optional[Callable[[Tuple[T1]], Optional[ErrType]]],
            Optional[Callable[[Tuple[T1, T2]], Optional[ErrType]]],
            Optional[Callable[[Tuple[T1, T2, T3]], Optional[ErrType]]],
            Optional[Callable[[Tuple[T1, T2, T3, T4]], Optional[ErrType]]],
            Optional[Callable[[Tuple[T1, T2, T3, T4, T5]], Optional[ErrType]]],
            Optional[Callable[[Tuple[T1, T2, T3, T4, T5, T6]], Optional[ErrType]]],
            Optional[Callable[[Tuple[T1, T2, T3, T4, T5, T6, T7]], Optional[ErrType]]],
            Optional[
                Callable[[Tuple[T1, T2, T3, T4, T5, T6, T7, T8]], Optional[ErrType]]
            ],
        ] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> Union[
        "NTupleValidator[Tuple[T1]]",
        "NTupleValidator[Tuple[T1, T2]]",
        "NTupleValidator[Tuple[T1, T2, T3]]",
        "NTupleValidator[Tuple[T1, T2, T3, T4]]",
        "NTupleValidator[Tuple[T1, T2, T3, T4, T5]]",
        "NTupleValidator[Tuple[T1, T2, T3, T4, T5, T6]]",
        "NTupleValidator[Tuple[T1, T2, T3, T4, T5, T6, T7]]",
        "NTupleValidator[Tuple[T1, T2, T3, T4, T5, T6, T7, T8]]",
    ]:
        """
        Can be used for up to 8 typed tuple slots. For more than 8 slots,
        use ``NTupleValidator.untyped``.

        :param fields: a tuple the validators to be used, whose types will reflect
            the resulting tuple
        :param validate_object: a function which can be used to validate the tuple after
            individual slots have been validated
        :param coerce: control coercion prior to validation
        :return: ``NTupleValidator`` with the ``fields`` as defined by the ``fields``
            parameter
        """
        return NTupleValidator(
            fields=fields, validate_object=validate_object, coerce=coerce
        )  # type: ignore

    @staticmethod
    def untyped(
        *,
        fields: Tuple[Validator[Any], ...],
        validate_object: Optional[Callable[[Tuple[Any, ...]], Optional[ErrType]]] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> "NTupleValidator[Tuple[Any, ...]]":
        """
        Identical to NTupleValidator.typed except that it can be used for an
        arbitrary number of slots. Does not retain type information.

        :param fields: a tuple the validators to be used, whose types will reflect
            the resulting tuple
        :param validate_object: a function which can be used to validate the tuple after
            individual slots have been validated
        :param coerce: control coercion prior to validation
        :return: ``NTupleValidator`` with the ``fields`` as defined by the ``fields``
            parameter
        """
        return NTupleValidator(
            fields=fields, validate_object=validate_object, coerce=coerce
        )

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[A]:
        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, list), val, self
                )
            else:
                coerced_val: Tuple[Any, ...] = coerced.val
        elif type(val) is tuple:
            coerced_val = val
        else:
            return False, Invalid(TypeErr(tuple), val, self)

        # we allow list as well because it's common that tuples or tuple-like lists
        # are deserialized to lists
        if not self._len_predicate(coerced_val):
            return False, Invalid(PredicateErrs([self._len_predicate]), coerced_val, self)
        errs: Dict[int, Invalid] = {}
        vals = []
        for i, (validator, tuple_val) in enumerate(
            zip(self._wrapped_fields_sync, coerced_val)
        ):
            succeeded, new_val = validator(tuple_val)
            if succeeded:
                vals.append(new_val)
            else:
                errs[i] = new_val

        if errs:
            return False, Invalid(IndexErrs(errs), coerced_val, self)
        else:
            obj: A = tuple(vals)  # type: ignore
            if self.validate_object is None:
                return True, obj
            else:
                obj_result = self.validate_object(obj)
                if obj_result is None:
                    return True, obj
                else:
                    return False, Invalid(obj_result, obj, self)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[A]:
        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, list), val, self
                )
            else:
                coerced_val: Tuple[Any, ...] = coerced.val
        elif type(val) is tuple:
            coerced_val = val
        else:
            return False, Invalid(TypeErr(tuple), val, self)

        if not self._len_predicate(val):
            return False, Invalid(PredicateErrs([self._len_predicate]), val, self)

        errs: Dict[int, Invalid] = {}
        vals = []
        for i, (validator, tuple_val) in enumerate(
            zip(self._wrapped_fields_async, coerced_val)
        ):
            succeeded, new_val = await validator(tuple_val)
            if succeeded:
                vals.append(new_val)
            else:
                errs[i] = new_val

        if errs:
            return False, Invalid(IndexErrs(errs), coerced_val, self)
        else:
            obj: A = tuple(vals)  # type: ignore
            if self.validate_object is None:
                return True, obj
            else:
                obj_result = self.validate_object(obj)
                if obj_result is None:
                    return True, obj
                else:
                    return False, Invalid(obj_result, obj, self)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.fields == other.fields
            and self.validate_object == other.validate_object
            and self.coerce == other.coerce
        )

    def __repr__(self) -> str:
        fields_str = "(" + ", ".join([repr(f) for f in self.fields]) + ")"
        if self.validate_object is not None:
            fields_str += f", validate_object={repr(self.validate_object)}"

        if self.coerce is not None:
            fields_str += f", coerce={repr(self.coerce)}"

        return f"{self.__class__.__name__}(fields={fields_str})"


class UniformTupleValidator(_ToTupleValidator[Tuple[A, ...]]):
    __match_args__ = ("item_validator", "predicates", "predicates_async", "coerce")

    def __init__(
        self,
        item_validator: Validator[A],
        *,
        predicates: Optional[List[Predicate[Tuple[A, ...]]]] = None,
        predicates_async: Optional[List[PredicateAsync[Tuple[A, ...]]]] = None,
        coerce: Optional[Coercer[Tuple[Any, ...]]] = tuple_or_list_to_tuple,
    ) -> None:
        self.item_validator = item_validator
        self.predicates = predicates
        self.predicates_async = predicates_async
        self.coerce = coerce

        self._item_validator_is_tuple = isinstance(item_validator, _ToTupleValidator)

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[Tuple[A, ...]]:
        if self.predicates_async:
            _async_predicates_warning(self.__class__)

        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, list), val, self
                )
            else:
                coerced_val: Tuple[Any, ...] = coerced.val

        elif type(val) is tuple:
            coerced_val = val
        else:
            return False, Invalid(TypeErr(tuple), val, self)

        if self.predicates:
            tuple_errors: List[
                Union[Predicate[Tuple[A, ...]], PredicateAsync[Tuple[A, ...]]]
            ] = [pred for pred in self.predicates if not pred(coerced_val)]

            # Not running async validators! They shouldn't be set!
            if tuple_errors:
                return False, Invalid(PredicateErrs(tuple_errors), coerced_val, self)

        return_list: List[A] = []
        index_errors: Dict[int, Invalid] = {}
        for i, item in enumerate(coerced_val):
            if self._item_validator_is_tuple:
                is_valid, item_result = self.item_validator._validate_to_tuple(item)  # type: ignore # noqa: E501
            else:
                _result = self.item_validator(item)
                is_valid, item_result = (
                    (True, _result.val) if _result.is_valid else (False, _result)
                )

            if not is_valid:
                index_errors[i] = item_result
            elif not index_errors:
                return_list.append(item_result)

        if index_errors:
            return False, Invalid(IndexErrs(index_errors), coerced_val, self)
        else:
            return True, tuple(return_list)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[Tuple[A, ...]]:
        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, list), val, self
                )
            else:
                coerced_val: Tuple[Any, ...] = coerced.val

        elif type(val) is tuple:
            coerced_val = val
        else:
            return False, Invalid(TypeErr(tuple), val, self)

        tuple_errors: List[
            Union[Predicate[Tuple[A, ...]], PredicateAsync[Tuple[A, ...]]]
        ] = []
        if self.predicates:
            tuple_errors.extend(
                [pred for pred in self.predicates if not pred(coerced_val)]
            )
        if self.predicates_async:
            for pred_async in self.predicates_async:
                if not await pred_async.validate_async(coerced_val):
                    tuple_errors.append(pred_async)

        # Not running async validators! They shouldn't be set!
        if tuple_errors:
            return False, Invalid(PredicateErrs(tuple_errors), coerced_val, self)

        return_list: List[A] = []
        index_errors: Dict[int, Invalid] = {}
        for i, item in enumerate(coerced_val):
            if self._item_validator_is_tuple:
                (
                    is_valid,
                    item_result,
                ) = await self.item_validator._validate_to_tuple_async(  # type: ignore # noqa: E501
                    item
                )
            else:
                _result = await self.item_validator.validate_async(item)
                is_valid, item_result = (
                    (True, _result.val) if _result.is_valid else (False, _result)
                )

            if not is_valid:
                index_errors[i] = item_result
            elif not index_errors:
                return_list.append(item_result)

        if index_errors:
            return False, Invalid(IndexErrs(index_errors), coerced_val, self)
        else:
            return True, tuple(return_list)

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) is type(self)
            and self.item_validator == other.item_validator
            and self.predicates == other.predicates
            and self.predicates_async == other.predicates_async
            and self.coerce == other.coerce
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [repr(self.item_validator)]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("predicates", self.predicates),
                    ("predicates_async", self.predicates_async),
                    ("coerce", self.coerce),
                ]
                if v
            ],
        )



import dataclasses
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Dict,
    Hashable,
    List,
    Optional,
    Tuple,
    Union,
    overload,
)

from koda import Just, Maybe, nothing

from koda_validate._generics import (
    T1,
    T2,
    T3,
    T4,
    T5,
    T6,
    T7,
    T8,
    T9,
    T10,
    T11,
    T12,
    T13,
    T14,
    T15,
    T16,
    A,
    Ret,
)
from koda_validate._internal import (
    _async_predicates_warning,
    _raise_cannot_define_validate_object_and_validate_object_async,
    _raise_validate_object_async_in_sync_mode,
    _repr_helper,
    _ResultTuple,
    _ToTupleValidator,
    _wrap_async_validator,
    _wrap_sync_validator,
)
from koda_validate.base import Predicate, PredicateAsync, Validator
from koda_validate.coerce import Coercer
from koda_validate.errors import (
    CoercionErr,
    ErrType,
    ExtraKeysErr,
    KeyErrs,
    KeyValErrs,
    MapErr,
    MissingKeyErr,
    PredicateErrs,
    TypeErr,
    missing_key_err,
)
from koda_validate.valid import Invalid, Valid, ValidationResult


class KeyNotRequired(Validator[Maybe[A]]):
    def __init__(self, validator: Validator[A]):
        self.validator = validator

    async def validate_async(self, val: Any) -> ValidationResult[Maybe[A]]:
        result = await self.validator.validate_async(val)
        if result.is_valid:
            return Valid(Just(result.val))
        else:
            return result

    def __call__(self, val: Any) -> ValidationResult[Maybe[A]]:
        result = self.validator(val)
        if result.is_valid:
            return Valid(Just(result.val))
        else:
            return result

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.validator == other.validator

    def __repr__(self) -> str:
        return f"KeyNotRequired({repr(self.validator)})"


KeyValidator = Tuple[
    Hashable,
    Validator[A],
]


class MapValidator(Validator[Dict[T1, T2]]):
    __match_args__ = (
        "key_validator",
        "value_validator",
        "predicates",
        "predicates_async",
        "coerce",
    )

    def __init__(
        self,
        *,
        key: Validator[T1],
        value: Validator[T2],
        predicates: Optional[List[Predicate[Dict[T1, T2]]]] = None,
        predicates_async: Optional[List[PredicateAsync[Dict[T1, T2]]]] = None,
        coerce: Optional[Coercer[Dict[Any, Any]]] = None,
    ) -> None:
        self.key_validator = key
        self.value_validator = value
        self.predicates = predicates
        self.predicates_async = predicates_async
        self.coerce = coerce

    async def validate_async(self, val: Any) -> ValidationResult[Dict[T1, T2]]:
        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return Invalid(CoercionErr(self.coerce.compatible_types, dict), val, self)
            else:
                coerced_val: Dict[Any, Any] = coerced.val

        elif type(val) is dict:
            coerced_val = val
        else:
            return Invalid(TypeErr(dict), val, self)

        predicate_errors: List[
            Union[Predicate[Dict[Any, Any]], PredicateAsync[Dict[Any, Any]]]
        ] = []
        if self.predicates is not None:
            for predicate in self.predicates:
                # Note that the expectation here is that validators will likely
                # be doing json like number of keys; they aren't expected
                # to be drilling down into specific keys and values. That may be
                # an incorrect assumption; if so, some minor refactoring is probably
                # necessary.
                if not predicate(coerced_val):
                    predicate_errors.append(predicate)

        if self.predicates_async is not None:
            for pred_async in self.predicates_async:
                if not await pred_async.validate_async(coerced_val):
                    predicate_errors.append(pred_async)

        if predicate_errors:
            return Invalid(PredicateErrs(predicate_errors), coerced_val, self)

        return_dict: Dict[T1, T2] = {}
        errors: Dict[Any, KeyValErrs] = {}

        for key, val_ in coerced_val.items():
            key_result = await self.key_validator.validate_async(key)
            val_result = await self.value_validator.validate_async(val_)

            if key_result.is_valid and val_result.is_valid:
                return_dict[key_result.val] = val_result.val
            else:
                errors[key] = KeyValErrs(
                    key=None if key_result.is_valid else key_result,
                    val=None if val_result.is_valid else val_result,
                )

        if errors:
            return Invalid(MapErr(errors), coerced_val, self)
        else:
            return Valid(return_dict)

    def __call__(self, val: Any) -> ValidationResult[Dict[T1, T2]]:
        if self.predicates_async:
            _async_predicates_warning(self.__class__)

        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return Invalid(CoercionErr(self.coerce.compatible_types, dict), val, self)
            else:
                coerced_val: Dict[Any, Any] = coerced.val

        elif type(val) is dict:
            coerced_val = val
        else:
            return Invalid(TypeErr(dict), val, self)

        predicate_errors: List[
            Union[Predicate[Dict[Any, Any]], PredicateAsync[Dict[Any, Any]]]
        ] = []
        if self.predicates is not None:
            for predicate in self.predicates:
                # Note that the expectation here is that validators will likely
                # be doing json like number of keys; they aren't expected
                # to be drilling down into specific keys and values. That may be
                # an incorrect assumption; if so, some minor refactoring is probably
                # necessary.
                if not predicate(coerced_val):
                    predicate_errors.append(predicate)

        if predicate_errors:
            return Invalid(PredicateErrs(predicate_errors), coerced_val, self)

        return_dict: Dict[T1, T2] = {}
        errors: Dict[Any, KeyValErrs] = {}
        for key, val_ in coerced_val.items():
            key_result = self.key_validator(key)
            val_result = self.value_validator(val_)

            if key_result.is_valid and val_result.is_valid:
                return_dict[key_result.val] = val_result.val
            else:
                errors[key] = KeyValErrs(
                    key=None if key_result.is_valid else key_result,
                    val=None if val_result.is_valid else val_result,
                )

        if errors:
            return Invalid(MapErr(errors), coerced_val, self)
        else:
            return Valid(return_dict)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.key_validator == other.key_validator
            and self.value_validator == other.value_validator
            and self.predicates == other.predicates
            and self.predicates_async == other.predicates_async
            and self.coerce == other.coerce
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [
                f"key={repr(self.key_validator)}",
                f"value={repr(self.value_validator)}",
            ]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("predicates", self.predicates),
                    ("predicates_async", self.predicates_async),
                    ("coerce", self.coerce),
                ]
                if v
            ],
        )


class IsDictValidator(_ToTupleValidator[Dict[Any, Any]]):
    _instance: ClassVar[Optional["IsDictValidator"]] = None

    def __new__(cls) -> "IsDictValidator":
        # make a singleton
        if cls._instance is None:
            cls._instance = super(IsDictValidator, cls).__new__(cls)
        return cls._instance

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[Dict[Any, Any]]:
        if isinstance(val, dict):
            return True, val
        else:
            return False, Invalid(TypeErr(dict), val, self)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[Dict[Any, Any]]:
        return self._validate_to_tuple(val)

    def __repr__(self) -> str:
        return "IsDictValidator()"


is_dict_validator = IsDictValidator()


@dataclasses.dataclass
class MinKeys(Predicate[Dict[Any, Any]]):
    size: int

    def __call__(self, val: Dict[Any, Any]) -> bool:
        return len(val) >= self.size


@dataclasses.dataclass
class MaxKeys(Predicate[Dict[Any, Any]]):
    size: int

    def __call__(self, val: Dict[Any, Any]) -> bool:
        return len(val) <= self.size


class RecordValidator(_ToTupleValidator[Ret]):
    __match_args__ = (
        "keys",
        "into",
        "validate_object",
        "validate_object_async",
    )

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1], Ret],
        keys: Tuple[
            KeyValidator[T1],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7, T8], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
            KeyValidator[T11],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
            KeyValidator[T11],
            KeyValidator[T12],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13], Ret],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
            KeyValidator[T11],
            KeyValidator[T12],
            KeyValidator[T13],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[
            [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14], Ret
        ],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
            KeyValidator[T11],
            KeyValidator[T12],
            KeyValidator[T13],
            KeyValidator[T14],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[
            [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15], Ret
        ],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
            KeyValidator[T11],
            KeyValidator[T12],
            KeyValidator[T13],
            KeyValidator[T14],
            KeyValidator[T15],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    @overload
    def __init__(
        self,
        *,
        into: Callable[
            [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16], Ret
        ],
        keys: Tuple[
            KeyValidator[T1],
            KeyValidator[T2],
            KeyValidator[T3],
            KeyValidator[T4],
            KeyValidator[T5],
            KeyValidator[T6],
            KeyValidator[T7],
            KeyValidator[T8],
            KeyValidator[T9],
            KeyValidator[T10],
            KeyValidator[T11],
            KeyValidator[T12],
            KeyValidator[T13],
            KeyValidator[T14],
            KeyValidator[T15],
            KeyValidator[T16],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        ...  # pragma: no cover

    def __init__(
        self,
        into: Union[
            Callable[[T1], Ret],
            Callable[[T1, T2], Ret],
            Callable[[T1, T2, T3], Ret],
            Callable[[T1, T2, T3, T4], Ret],
            Callable[[T1, T2, T3, T4, T5], Ret],
            Callable[[T1, T2, T3, T4, T5, T6], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13], Ret],
            Callable[[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14], Ret],
            Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15], Ret
            ],
            Callable[
                [T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16],
                Ret,
            ],
        ],
        keys: Union[
            Tuple[KeyValidator[T1]],
            Tuple[KeyValidator[T1], KeyValidator[T2]],
            Tuple[KeyValidator[T1], KeyValidator[T2], KeyValidator[T3]],
            Tuple[KeyValidator[T1], KeyValidator[T2], KeyValidator[T3], KeyValidator[T4]],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
                KeyValidator[T11],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
                KeyValidator[T11],
                KeyValidator[T12],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
                KeyValidator[T11],
                KeyValidator[T12],
                KeyValidator[T13],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
                KeyValidator[T11],
                KeyValidator[T12],
                KeyValidator[T13],
                KeyValidator[T14],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
                KeyValidator[T11],
                KeyValidator[T12],
                KeyValidator[T13],
                KeyValidator[T14],
                KeyValidator[T15],
            ],
            Tuple[
                KeyValidator[T1],
                KeyValidator[T2],
                KeyValidator[T3],
                KeyValidator[T4],
                KeyValidator[T5],
                KeyValidator[T6],
                KeyValidator[T7],
                KeyValidator[T8],
                KeyValidator[T9],
                KeyValidator[T10],
                KeyValidator[T11],
                KeyValidator[T12],
                KeyValidator[T13],
                KeyValidator[T14],
                KeyValidator[T15],
                KeyValidator[T16],
            ],
        ],
        validate_object: Optional[Callable[[Ret], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[Ret], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:

        self.into = into
        # needs to be `Any` until we have variadic generics presumably
        self.keys: Tuple[KeyValidator[Any], ...] = keys
        if validate_object is not None and validate_object_async is not None:
            _raise_cannot_define_validate_object_and_validate_object_async()
        self.validate_object = validate_object
        self.validate_object_async = validate_object_async
        self.fail_on_unknown_keys = fail_on_unknown_keys

        self._disallow_synchronous = bool(validate_object_async)

        # so we don't need to calculate each time we validate
        self._key_set = set()
        self._fast_keys_sync: List[
            Tuple[Hashable, Callable[[Any], _ResultTuple[Any]], bool]
        ] = []

        self._fast_keys_async: List[
            Tuple[Hashable, Callable[[Any], Awaitable[_ResultTuple[Any]]], bool]
        ] = []

        for key, val in keys:
            is_required = not isinstance(val, KeyNotRequired)
            self._fast_keys_sync.append((key, _wrap_sync_validator(val), is_required))  # type: ignore  # noqa: E501
            self._fast_keys_async.append((key, _wrap_async_validator(val), is_required))  # type: ignore  # noqa: E501
            self._key_set.add(key)

        self._unknown_keys_err: ExtraKeysErr = ExtraKeysErr(self._key_set)

    def _validate_to_tuple(self, data: Any) -> _ResultTuple[Ret]:
        if self._disallow_synchronous:
            _raise_validate_object_async_in_sync_mode(self.__class__)
        if not isinstance(data, dict):
            return False, Invalid(TypeErr(dict), data, self)

        if self.fail_on_unknown_keys:
            for key_ in data:
                if key_ not in self._key_set:
                    return False, Invalid(self._unknown_keys_err, data, self)

        args: List[Any] = []
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_sync:
            if key_ not in data:
                if key_required:
                    errs[key_] = Invalid(MissingKeyErr(), data, self)
                elif not errs:
                    args.append(nothing)
            else:
                success, new_val = validator(data[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    args.append(new_val)

        if errs:
            return False, Invalid(KeyErrs(errs), data, self)
        else:
            obj = self.into(*args)
            if self.validate_object and (result := self.validate_object(obj)):
                return False, Invalid(result, obj, self)
            return True, obj

    async def _validate_to_tuple_async(self, data: Any) -> _ResultTuple[Ret]:
        if not isinstance(data, dict):
            return False, Invalid(TypeErr(dict), data, self)

        if self.fail_on_unknown_keys:
            for key_ in data:
                if key_ not in self._key_set:
                    return False, Invalid(self._unknown_keys_err, data, self)

        args: List[Any] = []
        errs: Dict[Any, Invalid] = {}
        for key_, async_validator, key_required in self._fast_keys_async:
            if key_ not in data:
                if key_required:
                    errs[key_] = Invalid(MissingKeyErr(), data, self)
                else:
                    args.append(nothing)
            else:
                success, new_val = await async_validator(data[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    args.append(new_val)

        if errs:
            return False, Invalid(KeyErrs(errs), data, self)
        else:
            obj = self.into(*args)
            if self.validate_object and (result := self.validate_object(obj)):
                return False, Invalid(result, obj, self)
            elif self.validate_object_async and (
                async_result := await self.validate_object_async(obj)
            ):
                return False, Invalid(async_result, obj, self)

            return True, obj

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.into == other.into
            and self.keys == other.keys
            and self.validate_object == other.validate_object
            and self.validate_object_async == other.validate_object_async
            and self.fail_on_unknown_keys == other.fail_on_unknown_keys
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [f"keys={repr(self.keys)}", f"into={repr(self.into)}"]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("validate_object", self.validate_object),
                    ("validate_object_async", self.validate_object_async),
                    # note that this coincidentally works as we want:
                    # by default we don't fail on extra keys, so we don't
                    # show this in the repr if the default is defined
                    ("fail_on_unknown_keys", self.fail_on_unknown_keys),
                ]
                if v
            ],
        )


class DictValidatorAny(_ToTupleValidator[Dict[Any, Any]]):
    """
    This class exists for a few reasons:

    - it can handle an arbitrary amount of keys, of any combination of hashable types

    - it's initialization is very straightforward

    The big caveat is that a valid result is always typed as ``Dict[Any, Any]``, even
    though the validation will work just as well as with any other :class:`Validator`

    """

    __match_args__ = (
        "schema",
        "validate_object",
        "validate_object_async",
    )

    def __init__(
        self,
        schema: Dict[Any, Validator[Any]],
        *,
        validate_object: Optional[
            Callable[[Dict[Hashable, Any]], Optional[ErrType]]
        ] = None,
        validate_object_async: Optional[
            Callable[
                [Dict[Any, Any]],
                Awaitable[Optional[ErrType]],
            ]
        ] = None,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        self.schema: Dict[Any, Validator[Any]] = schema
        self.validate_object = validate_object
        self.validate_object_async = validate_object_async

        if validate_object is not None and validate_object_async is not None:
            _raise_cannot_define_validate_object_and_validate_object_async()
        self.fail_on_unknown_keys = fail_on_unknown_keys

        self._disallow_synchronous = bool(validate_object_async)

        # so we don't need to calculate each time we validate
        self._fast_keys_sync = []
        self._fast_keys_async = []
        self._keys_set = set()
        for key, val in schema.items():
            self._keys_set.add(key)
            vldtr = (
                val.validator
                if (is_not_required := isinstance(val, KeyNotRequired))
                else val
            )
            self._fast_keys_sync.append(
                (key, _wrap_sync_validator(vldtr), not is_not_required)
            )
            self._fast_keys_async.append(
                (key, _wrap_async_validator(vldtr), not is_not_required)
            )

        self._unknown_keys_err = ExtraKeysErr(set(schema.keys()))

    def _validate_to_tuple(self, data: Any) -> _ResultTuple[Dict[Any, Any]]:
        if self._disallow_synchronous:
            _raise_validate_object_async_in_sync_mode(self.__class__)

        if not type(data) is dict:
            return False, Invalid(TypeErr(dict), data, self)

        if self.fail_on_unknown_keys:
            for key_ in data:
                if key_ not in self._keys_set:
                    return False, Invalid(self._unknown_keys_err, data, self)

        success_dict: Dict[Any, Any] = {}
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_sync:
            if key_ not in data:
                if key_required:
                    errs[key_] = Invalid(missing_key_err, data, self)
            else:
                success, new_val = validator(data[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    success_dict[key_] = new_val

        if errs:
            return False, Invalid(KeyErrs(errs), data, self)
        elif self.validate_object and (result := self.validate_object(success_dict)):
            return False, Invalid(result, success_dict, self)

        return True, success_dict

    async def _validate_to_tuple_async(self, data: Any) -> _ResultTuple[Dict[Any, Any]]:
        if not type(data) is dict:
            return False, Invalid(TypeErr(dict), data, self)

        if self.fail_on_unknown_keys:
            # this seems to be faster than `for key_ in data.keys()`
            for key_ in data:
                if key_ not in self._keys_set:
                    return False, Invalid(self._unknown_keys_err, data, self)

        success_dict: Dict[Any, Any] = {}
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_async:
            if key_ not in data:
                if key_required:
                    errs[key_] = Invalid(missing_key_err, data, self)
            else:
                success, new_val = await validator(data[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    success_dict[key_] = new_val

        if errs:
            return False, Invalid(KeyErrs(errs), data, self)
        elif self.validate_object and (result := self.validate_object(success_dict)):
            return False, Invalid(result, success_dict, self)
        elif self.validate_object_async and (
            result := await self.validate_object_async(success_dict)
        ):
            return False, Invalid(result, success_dict, self)

        return True, success_dict

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.schema == other.schema
            and self.validate_object == other.validate_object
            and self.validate_object_async == other.validate_object_async
            and self.fail_on_unknown_keys == other.fail_on_unknown_keys
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [repr(self.schema)]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("validate_object", self.validate_object),
                    ("validate_object_async", self.validate_object_async),
                    # note that this coincidentally works as we want:
                    # by default we don't fail on extra keys, so we don't
                    # show this in the repr if the default is defined
                    ("fail_on_unknown_keys", self.fail_on_unknown_keys),
                ]
                if v
            ],
        )



import sys
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    FrozenSet,
    Mapping,
    Optional,
    Type,
    TypeVar,
    cast,
    get_type_hints,
)

from koda_validate._internal import (
    _is_typed_dict_cls,
    _raise_cannot_define_validate_object_and_validate_object_async,
    _raise_validate_object_async_in_sync_mode,
    _repr_helper,
    _ResultTuple,
    _ToTupleValidator,
    _wrap_async_validator,
    _wrap_sync_validator,
)
from koda_validate.base import Validator
from koda_validate.coerce import Coercer
from koda_validate.errors import (
    CoercionErr,
    ErrType,
    ExtraKeysErr,
    KeyErrs,
    TypeErr,
    missing_key_err,
)
from koda_validate.typehints import get_typehint_validator
from koda_validate.valid import Invalid

_TDT = TypeVar("_TDT", bound=Mapping[str, object])


class TypedDictValidator(_ToTupleValidator[_TDT]):
    """
    Takes a ``TypedDict`` subclass as an argument and derives a :class:`Validator`.

    Optional keys are determined by the ``__optional_keys__`` and ``__required_keys__``
    attributes.

    .. note::

        This validator _might_ work on non-typed-dict classes (There are challenges
        in defining a TypedDict-like type). *Please* do not intentionally try to use
        it for non-TypedDict datatypes.

    Example:

    .. testcode:: tdexample

        from typing import List, TypedDict
        from koda_validate import *

        class Person(TypedDict):
            name: str
            hobbies: List[str]

        validator = TypedDictValidator(Person)

    Usage:

    .. doctest:: tdexample

        >>> validator({"name": "Bob", "hobbies": ["eating", "coding", "sleeping"]})
        Valid(val={'name': 'Bob', 'hobbies': ['eating', 'coding', 'sleeping']})


    :param td_cls: A ``TypedDict`` subclass
    :param overrides: a dict whose keys define explicit validators
    :param validate_object: is run if all keys have been validated individually. If it
        returns ``None``, then there were no errors; otherwise it should return
        ``ErrType``
    :param validate_object_async: same as ``validate_object``, except that is runs
        asynchronously
    :param typehint_resolver: define this to override default inferred validators for
        types
    :param coerce: this can be set to create any kind of custom coercion
    :param fail_on_unknown_keys: if True, this will fail if any keys not defined by the
        ``TypedDict`` are found. This will fail before any values are validated.
    :raises TypeError: should raise if non-``TypedDict`` type is passed for ``td_cls``
    """

    __match_args__ = ("td_cls", "overrides", "fail_on_unknown_keys", "coerce")

    def __init__(
        self,
        td_cls: Type[_TDT],
        *,
        overrides: Optional[Dict[str, Validator[Any]]] = None,
        validate_object: Optional[Callable[[_TDT], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[
                [_TDT],
                Awaitable[Optional[ErrType]],
            ]
        ] = None,
        coerce: Optional[Coercer[Dict[Any, Any]]] = None,
        typehint_resolver: Callable[[Any], Validator[Any]] = get_typehint_validator,
        fail_on_unknown_keys: bool = False,
    ) -> None:
        if not _is_typed_dict_cls(td_cls):
            raise TypeError("must be a TypedDict subclass")

        self.td_cls = td_cls
        self.overrides = overrides  # for repr
        self.fail_on_unknown_keys = fail_on_unknown_keys
        self.coerce = coerce

        if validate_object is not None and validate_object_async is not None:
            _raise_cannot_define_validate_object_and_validate_object_async()

        self.validate_object = validate_object
        self.validate_object_async = validate_object_async

        self._disallow_synchronous = bool(validate_object_async)

        if sys.version_info >= (3, 9):
            # Required/NotRequired keys are always present in
            self.required_keys: FrozenSet[str] = getattr(
                td_cls, "__required_keys__", frozenset()
            )
            type_hints = get_type_hints(self.td_cls, include_extras=True)
        else:
            # not going to try to handle superclasses
            self.required_keys = (
                frozenset([k for k in td_cls.__annotations__])
                if getattr(td_cls, "__total__", True)
                else frozenset()
            )
            type_hints = get_type_hints(self.td_cls)

        overrides = self.overrides or {}
        self.schema = {
            field: (
                overrides[field] if field in overrides else typehint_resolver(annotations)
            )
            for field, annotations in type_hints.items()
        }

        self._keys_set = set()
        self._fast_keys_sync = []
        self._fast_keys_async = []
        for key, val in self.schema.items():
            self._keys_set.add(key)
            is_required = key in self.required_keys
            self._fast_keys_sync.append((key, _wrap_sync_validator(val), is_required))
            self._fast_keys_async.append((key, _wrap_async_validator(val), is_required))

        self._unknown_keys_err: ExtraKeysErr = ExtraKeysErr(set(self.schema.keys()))

    def _validate_to_tuple(self, data: Any) -> _ResultTuple[_TDT]:
        if self._disallow_synchronous:
            _raise_validate_object_async_in_sync_mode(self.__class__)

        if self.coerce:
            if not (coerced := self.coerce(data)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, dict), data, self
                )
            else:
                coerced_val: Dict[Any, Any] = coerced.val

        elif type(data) is dict:
            coerced_val = data
        else:
            return False, Invalid(TypeErr(dict), data, self)

        if self.fail_on_unknown_keys:
            for key_ in coerced_val:
                if key_ not in self._keys_set:
                    return False, Invalid(self._unknown_keys_err, coerced_val, self)

        success_dict: Dict[str, object] = {}
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_sync:
            if key_ not in coerced_val:
                if key_required:
                    errs[key_] = Invalid(missing_key_err, coerced_val, self)
            else:
                success, new_val = validator(coerced_val[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    success_dict[key_] = new_val

        if errs:
            return False, Invalid(KeyErrs(errs), coerced_val, self)
        else:
            if self.validate_object and (
                result := self.validate_object(cast(_TDT, success_dict))
            ):
                return False, Invalid(result, success_dict, self)
            return True, cast(_TDT, success_dict)

    async def _validate_to_tuple_async(self, data: Any) -> _ResultTuple[_TDT]:
        if self.coerce:
            if not (coerced := self.coerce(data)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, dict), data, self
                )
            else:
                coerced_val: Dict[Any, Any] = coerced.val

        elif type(data) is dict:
            coerced_val = data
        else:
            return False, Invalid(TypeErr(dict), data, self)

        if self.fail_on_unknown_keys:
            for key_ in coerced_val:
                if key_ not in self._keys_set:
                    return False, Invalid(self._unknown_keys_err, data, self)

        success_dict: Dict[str, object] = {}
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_async:
            if key_ not in coerced_val:
                if key_required:
                    errs[key_] = Invalid(missing_key_err, coerced_val, self)
            else:
                success, new_val = await validator(coerced_val[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    success_dict[key_] = new_val

        if errs:
            return False, Invalid(KeyErrs(errs), coerced_val, self)
        else:
            if self.validate_object and (
                result := self.validate_object(cast(_TDT, success_dict))
            ):
                return False, Invalid(result, success_dict, self)
            elif self.validate_object_async and (
                result_async := await self.validate_object_async(cast(_TDT, success_dict))
            ):
                return False, Invalid(result_async, success_dict, self)
            return True, cast(_TDT, success_dict)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.schema == other.schema
            and self.validate_object == other.validate_object
            and self.validate_object_async == other.validate_object_async
            and self.fail_on_unknown_keys == other.fail_on_unknown_keys
            and self.coerce == other.coerce
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [repr(self.td_cls)]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("overrides", self.overrides),
                    ("validate_object", self.validate_object),
                    ("validate_object_async", self.validate_object_async),
                    ("coerce", self.coerce),
                    # note that this coincidentally works as we want:
                    # by default we don't fail on extra keys, so we don't
                    # show this in the repr if the default is defined
                    ("fail_on_unknown_keys", self.fail_on_unknown_keys),
                ]
                if v
            ],
        )



from typing import Any, List, Optional, Tuple, Union, overload

from koda_validate._generics import T1, T2, T3, T4, T5, T6, T7, T8, A
from koda_validate._internal import (
    _repr_helper,
    _ResultTuple,
    _ToTupleValidator,
    _union_validator,
    _union_validator_async,
)
from koda_validate.base import Validator


class UnionValidator(_ToTupleValidator[A]):
    __match_args__ = ("validators",)

    def __init__(
        self,
        # just a way to enforce at least one validator is defined
        validator_1: Validator[Any],
        *validators: Validator[Any],
    ) -> None:
        """
        You probably do not want to call the ``__init__`` method directly. Instead,
        look to UnionValidator.typed and UnionValidator.untyped. They are thin wrappers
        around the `__init__` method, but they allow more explicit control over types.

        :param validator_1: the first variant of the union. This is specified on its own,
            so we have at least one variant specified
        :param validators: the rest of the variants of the union
        """
        self.validators: Tuple[Validator[Any], ...] = (validator_1,) + validators

    @overload
    @staticmethod
    def typed(validator_1: Validator[T1], /) -> "UnionValidator[T1]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        /,
    ) -> "UnionValidator[Union[T1, T2]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        validator_3: Validator[T3],
        /,
    ) -> "UnionValidator[Union[T1, T2, T3]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        validator_3: Validator[T3],
        validator_4: Validator[T4],
        /,
    ) -> "UnionValidator[Union[T1, T2, T3, T4]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        validator_3: Validator[T3],
        validator_4: Validator[T4],
        validator_5: Validator[T5],
        /,
    ) -> "UnionValidator[Union[T1, T2, T3, T4, T5]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        validator_3: Validator[T3],
        validator_4: Validator[T4],
        validator_5: Validator[T5],
        validator_6: Validator[T6],
        /,
    ) -> "UnionValidator[Union[T1, T2, T3, T4, T5, T6]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        validator_3: Validator[T3],
        validator_4: Validator[T4],
        validator_5: Validator[T5],
        validator_6: Validator[T6],
        validator_7: Validator[T7],
        /,
    ) -> "UnionValidator[Union[T1, T2, T3, T4, T5, T6, T7]]":
        ...  # pragma: no cover

    @overload
    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Validator[T2],
        validator_3: Validator[T3],
        validator_4: Validator[T4],
        validator_5: Validator[T5],
        validator_6: Validator[T6],
        validator_7: Validator[T7],
        validator_8: Validator[T8],
        /,
    ) -> "UnionValidator[Union[T1, T2, T3, T4, T5, T6, T7, T8]]":
        ...  # pragma: no cover

    @staticmethod
    def typed(
        validator_1: Validator[T1],
        validator_2: Optional[Validator[T2]] = None,
        validator_3: Optional[Validator[T3]] = None,
        validator_4: Optional[Validator[T4]] = None,
        validator_5: Optional[Validator[T5]] = None,
        validator_6: Optional[Validator[T6]] = None,
        validator_7: Optional[Validator[T7]] = None,
        validator_8: Optional[Validator[T8]] = None,
    ) -> Union[
        "UnionValidator[T1]",
        "UnionValidator[Union[T1, T2]]",
        "UnionValidator[Union[T1, T2, T3]]",
        "UnionValidator[Union[T1, T2, T3, T4]]",
        "UnionValidator[Union[T1, T2, T3, T4, T5]]",
        "UnionValidator[Union[T1, T2, T3, T4, T5, T6]]",
        "UnionValidator[Union[T1, T2, T3, T4, T5, T6, T7]]",
        "UnionValidator[Union[T1, T2, T3, T4, T5, T6, T7, T8]]",
    ]:
        """
        Can be used for up to 8 typed variants. For more than 8 variants,
        use ``UnionValidator.untyped``. Arguments should be positional only.
        (``@overloads`` specify that arguments should be positional-only.)

        :param validator_1: the first variant
        :param validator_2: the second variant (if defined)
        :param validator_3: the third variant (if defined)
        :param validator_4: the fourth variant (if defined)
        :param validator_5: the fifth variant (if defined)
        :param validator_6: the sixth variant (if defined)
        :param validator_7: the seventh variant (if defined)
        :param validator_8: the eighth variant (if defined)
        :return: UnionValidator with the variants defined above.
        """
        validators: List[Validator[Any]] = [
            v
            for v in [
                validator_2,
                validator_3,
                validator_4,
                validator_5,
                validator_6,
                validator_7,
                validator_8,
            ]
            if v
        ]
        return UnionValidator(validator_1, *validators)

    @staticmethod
    def untyped(
        validator_1: Validator[Any], *validators: Validator[Any]
    ) -> "UnionValidator[Any]":
        r"""
        Can handle any number of variant :class:`Validator<koda_validate.Validator>``\s
        (as opposed to ``UnionValidator.typed``), but does not retain type information.

        :param validator_1: the first variant of the union. This is specified on its own,
            so we have at least one variant specified
        :param validators: the rest of the variants of the union
        :return: UnionValidator with the variants defined above.
        """
        return UnionValidator(validator_1, *validators)

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[A]:
        return _union_validator(self, self.validators, val)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[A]:
        return await _union_validator_async(self, self.validators, val)

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.validators == other.validators

    def __repr__(self) -> str:
        return _repr_helper(self.__class__, [repr(v) for v in self.validators])



from dataclasses import dataclass
from typing import Any, Callable, Generic, Set, Type

from koda import Maybe

from koda_validate._generics import A


@dataclass
class Coercer(Generic[A]):
    coerce: Callable[[Any], Maybe[A]]
    """
    The function which handles the coercion.
    """
    compatible_types: Set[Type[Any]]
    """
    All the types which can potentially be coerced.
    """

    def __call__(self, val: Any) -> Maybe[A]:
        return self.coerce(val)


def coercer(
    *compatible_types: Type[Any],
) -> Callable[[Callable[[Any], Maybe[A]]], Coercer[A]]:
    """
    This is purely a convenience constructor for :class:`Coercer` objects.

    :param compatible_types: the types the coercer can take to produce an
        the return type
    :return: A callable which accepts a function that should be congruent
        with the ``compatible_types`` param.

    """

    def inner(func: Callable[[Any], Maybe[A]]) -> Coercer[A]:
        return Coercer(func, set(compatible_types))

    return inner



from typing import List, Optional, Type

from koda_validate import Predicate, PredicateAsync, Processor
from koda_validate._generics import SuccessT
from koda_validate._internal import _ToTupleStandardValidator
from koda_validate.coerce import Coercer


class TypeValidator(_ToTupleStandardValidator[SuccessT]):
    def __init__(
        self,
        type_: Type[SuccessT],
        *,
        predicates: Optional[List[Predicate[SuccessT]]] = None,
        predicates_async: Optional[List[PredicateAsync[SuccessT]]] = None,
        preprocessors: Optional[List[Processor[SuccessT]]] = None,
        coerce: Optional[Coercer[SuccessT]] = None,
    ) -> None:
        self._TYPE = type_
        super().__init__(
            *(predicates or []),
            predicates_async=predicates_async,
            preprocessors=preprocessors,
            coerce=coerce,
        )



"""
Generics are defined here to avoid repetition throughout the library
"""

from typing import TypeVar

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")
T10 = TypeVar("T10")
T11 = TypeVar("T11")
T12 = TypeVar("T12")
T13 = TypeVar("T13")
T14 = TypeVar("T14")
T15 = TypeVar("T15")
T16 = TypeVar("T16")

Ret = TypeVar("Ret")

SuccessT = TypeVar("SuccessT")



from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    Hashable,
    List,
    Optional,
    Set,
    Type,
    Union,
)

from koda_validate._generics import A
from koda_validate.base import Predicate, PredicateAsync

if TYPE_CHECKING:
    from koda_validate.valid import Invalid


@dataclass
class CoercionErr:
    """
    Similar to a TypeErr, but when one or more types can be
    coerced to a destination type
    """

    compatible_types: Set[Type[Any]]
    dest_type: Type[Any]


@dataclass
class ContainerErr:
    """
    This is for simple containers like `Maybe` or `Result`
    """

    child: "Invalid"


@dataclass
class ExtraKeysErr:
    """
    extra keys were present in a dictionary
    """

    expected_keys: Set[Hashable]


@dataclass
class KeyErrs:
    """
    validation failures for key/value pairs on a record-like
    dictionary
    """

    keys: Dict[Any, "Invalid"]


@dataclass
class KeyValErrs:
    """
    Key and/or value errors from a single key/value pair. This
    is useful for mapping collections.
    """

    key: Optional["Invalid"]
    val: Optional["Invalid"]


@dataclass
class MapErr:
    """
    errors from key/value pairs of a map-like dictionary
    """

    keys: Dict[Any, KeyValErrs]


class MissingKeyErr:
    """
    A key is missing from a dictionary
    """

    _instance: ClassVar[Optional["MissingKeyErr"]] = None

    def __new__(cls) -> "MissingKeyErr":
        # make a singleton
        if cls._instance is None:
            cls._instance = super(MissingKeyErr, cls).__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"


missing_key_err = MissingKeyErr()


@dataclass
class IndexErrs:
    """
    dictionary of validation errors by index
    """

    indexes: Dict[int, "Invalid"]


@dataclass
class SetErrs:
    """
    Errors from items in a set.
    """

    item_errs: List["Invalid"]


@dataclass
class UnionErrs:
    """
    Errors from each variant of a union.
    """

    variants: List["Invalid"]


@dataclass
class PredicateErrs(Generic[A]):
    """
    A grouping of failed Predicates
    """

    predicates: List[Union["Predicate[A]", "PredicateAsync[A]"]]


class ValidationErrBase:
    """
    This class exists only to provide a class to subclass
    for custom error types
    """

    pass


@dataclass
class TypeErr:
    """
    A specific type was required but not found
    """

    expected_type: Type[Any]


ErrType = Union[
    CoercionErr,
    ContainerErr,
    ExtraKeysErr,
    IndexErrs,
    KeyErrs,
    MapErr,
    MissingKeyErr,
    # This seems like a type exception worth making..., but that might change in the
    # future. This is backwards compatible with existing code
    PredicateErrs[Any],
    SetErrs,
    TypeErr,
    ValidationErrBase,
    UnionErrs,
]



import re
from datetime import date, datetime
from decimal import Decimal
from functools import partial
from typing import Any, Callable, Dict, List, NoReturn, Type, Union
from uuid import UUID

from koda_validate import NotBlank, UUIDValidator
from koda_validate.base import CacheValidatorBase, Predicate, PredicateAsync, Validator
from koda_validate.boolean import BoolValidator
from koda_validate.bytes import BytesValidator
from koda_validate.dataclasses import DataclassValidator
from koda_validate.decimal import DecimalValidator
from koda_validate.dictionary import (
    DictValidatorAny,
    IsDictValidator,
    KeyNotRequired,
    MapValidator,
    MaxKeys,
    MinKeys,
    RecordValidator,
)
from koda_validate.float import FloatValidator
from koda_validate.generic import (
    Choices,
    EndsWith,
    EqualsValidator,
    EqualTo,
    ExactLength,
    Lazy,
    Max,
    MaxItems,
    MaxLength,
    Min,
    MinItems,
    MinLength,
    StartsWith,
    UniqueItems,
)
from koda_validate.integer import IntValidator
from koda_validate.list import ListValidator
from koda_validate.namedtuple import NamedTupleValidator
from koda_validate.none import OptionalValidator
from koda_validate.serialization.base import Serializable
from koda_validate.string import EmailPredicate, RegexPredicate, StringValidator
from koda_validate.time import DatetimeValidator, DateValidator
from koda_validate.tuple import NTupleValidator, UniformTupleValidator
from koda_validate.typeddict import TypedDictValidator
from koda_validate.union import UnionValidator

AnyValidatorOrPredicate = Union[Validator[Any], Predicate[Any], PredicateAsync[Any]]
ValidatorToSchema = Callable[[AnyValidatorOrPredicate], Dict[str, Serializable]]


def unhandled_type(obj: Any) -> NoReturn:
    raise TypeError(
        f"type {type(obj)} not handled. You may want to write a wrapper function."
    )


def get_base(t: Type[Any]) -> Dict[str, Any]:
    if t is str:
        return {"type": "string"}
    elif t is bytes:
        return {"type": "string", "format": "byte"}
    elif t is int:
        return {"type": "integer"}
    elif t is Decimal:
        return {
            "type": "string",
            "format": "number",
            # based on BigDecimal (Java) but without exponent support. At least one
            # digit required, before or after comma
            # valid values:  "100.234567", "010", "-.05", "+1", "10", "100."
            "pattern": r"^(\-|\+)?((\d+(\.\d*)?)|(\.\d+))$",
        }
    elif t is float:
        return {"type": "number"}
    elif t is date:
        return {"type": "string", "format": "date"}
    elif t is datetime:
        return {"type": "string", "format": "date-time"}
    elif t is bool:
        return {"type": "boolean"}
    elif t is UUID:
        return {"type": "string", "format": "uuid"}
    raise TypeError(f"Got unexpected type: {t}")


def string_schema(
    to_schema_fn: ValidatorToSchema, validator: StringValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(str)

    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))

    return ret


def bytes_schema(
    to_schema_fn: ValidatorToSchema, validator: BytesValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(bytes)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))

    return ret


def integer_schema(
    to_schema_fn: ValidatorToSchema, validator: IntValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(int)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))

    return ret


def decimal_schema(
    to_schema_fn: ValidatorToSchema, validator: DecimalValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(Decimal)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))
    return ret


def float_schema(
    to_schema_fn: ValidatorToSchema, validator: FloatValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(float)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))
    return ret


def date_schema(
    to_schema_fn: ValidatorToSchema, validator: DateValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(date)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))
    return ret


def datetime_schema(
    to_schema_fn: ValidatorToSchema, validator: DatetimeValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(datetime)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))
    return ret


def equals_schema(
    to_schema_fn: ValidatorToSchema, validator: EqualsValidator[Any]
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(type(validator.match))
    ret.update(to_schema_fn(validator.predicate))
    return ret


def boolean_schema(
    to_schema_fn: ValidatorToSchema, validator: BoolValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(bool)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))
    return ret


def uuid_schema(
    to_schema_fn: ValidatorToSchema, validator: UUIDValidator
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = get_base(UUID)
    for pred in list(validator.predicates) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))
    return ret


def array_of_schema(
    to_schema_fn: ValidatorToSchema,
    validator: Union[ListValidator[Any], UniformTupleValidator[Any]],
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = {
        "type": "array",
        "items": to_schema_fn(validator.item_validator),
    }

    for pred in (validator.predicates or []) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))

    return ret


def obj_schema(
    to_schema_fn: ValidatorToSchema,
    obj: RecordValidator[Any],
) -> Dict[str, Serializable]:
    required: List[str] = []
    properties: Dict[str, Serializable] = {}
    for label, field in obj.keys:
        str_label = str(label)
        if not isinstance(field, KeyNotRequired):
            required.append(str_label)

        properties[str_label] = to_schema_fn(field)
    return {
        "type": "object",
        "additionalProperties": not obj.fail_on_unknown_keys,
        "required": required,  # type: ignore
        "properties": properties,
    }


def dict_validator_schema(
    to_schema_fn: ValidatorToSchema,
    obj: DictValidatorAny,
) -> Dict[str, Serializable]:
    required: List[str] = []
    properties: Dict[str, Serializable] = {}
    for label, field in obj.schema.items():
        str_label = str(label)
        if not isinstance(field, KeyNotRequired):
            required.append(str_label)

        properties[str_label] = to_schema_fn(field)
    return {
        "type": "object",
        "additionalProperties": not obj.fail_on_unknown_keys,
        "required": required,  # type: ignore
        "properties": properties,
    }


def dataclass_validator_schema(
    to_schema_fn: ValidatorToSchema,
    obj: DataclassValidator[Any],
) -> Dict[str, Serializable]:
    properties: Dict[str, Serializable] = {}
    for label, field in obj.schema.items():
        str_label = str(label)
        properties[str_label] = to_schema_fn(field)
    return {
        "type": "object",
        "additionalProperties": not obj.fail_on_unknown_keys,
        "required": obj.required_fields,  # type: ignore
        "properties": properties,
    }


def namedtuple_validator_schema(
    to_schema_fn: ValidatorToSchema,
    obj: NamedTupleValidator[Any],
) -> Dict[str, Serializable]:
    properties: Dict[str, Serializable] = {}
    for label, field in obj.schema.items():
        str_label = str(label)
        properties[str_label] = to_schema_fn(field)
    return {
        "type": "object",
        "additionalProperties": not obj.fail_on_unknown_keys,
        "required": obj.required_fields,  # type: ignore
        "properties": properties,
    }


def typeddict_validator_schema(
    to_schema_fn: ValidatorToSchema,
    obj: TypedDictValidator[Any],
) -> Dict[str, Serializable]:
    properties: Dict[str, Serializable] = {}
    for label, field in obj.schema.items():
        str_label = str(label)
        properties[str_label] = to_schema_fn(field)
    return {
        "type": "object",
        "additionalProperties": not obj.fail_on_unknown_keys,
        "required": sorted(obj.required_keys),  # type: ignore
        "properties": properties,
    }


def map_of_schema(
    to_schema_fn: ValidatorToSchema, validator: MapValidator[Any, Any]
) -> Dict[str, Serializable]:
    ret: Dict[str, Serializable] = {
        "type": "object",
        "additionalProperties": to_schema_fn(validator.value_validator),
    }

    for pred in (validator.predicates or []) + (validator.predicates_async or []):
        ret.update(to_schema_fn(pred))

    return ret


def generate_schema_predicate(
    pred: Union[Predicate[Any], PredicateAsync[Any]]
) -> Dict[str, Serializable]:
    # strings
    if isinstance(pred, EmailPredicate):
        return {"format": "email"}
    elif isinstance(pred, MaxLength):
        return {"maxLength": pred.length}
    elif isinstance(pred, MinLength):
        return {"minLength": pred.length}
    elif isinstance(pred, ExactLength):
        return {"minLength": pred.length, "maxLength": pred.length}
    elif isinstance(pred, Choices):
        return {"enum": (list(sorted(pred.choices)))}
    elif isinstance(pred, NotBlank):
        return {"pattern": r"^(?!\s*$).+"}
    elif isinstance(pred, RegexPredicate):
        return {"pattern": pred.pattern.pattern}
    elif isinstance(pred, StartsWith):
        return {"pattern": rf"^{re.escape(pred.prefix)}"}
    elif isinstance(pred, EndsWith):
        return {"pattern": rf"{re.escape(pred.suffix)}$"}
    # numbers
    elif isinstance(pred, Min):
        min_t = type(pred.minimum)
        if min_t is Decimal:
            min_ = str(pred.minimum)
        elif min_t is date or min_t is datetime:
            min_ = pred.minimum.isoformat()
        else:
            min_ = pred.minimum
        # non-standard format min / max
        if min_t in {Decimal, date, datetime}:
            return (
                {"formatExclusiveMinimum": min_}
                if pred.exclusive_minimum
                else {"formatMinimum": min_}
            )
        else:
            return (
                {"exclusiveMinimum": min_}
                if pred.exclusive_minimum
                else {"minimum": min_}
            )
    elif isinstance(pred, Max):
        max_t = type(pred.maximum)
        if max_t is Decimal:
            max_ = str(pred.maximum)
        elif max_t is date or max_t is datetime:
            max_ = pred.maximum.isoformat()
        else:
            max_ = pred.maximum

        # non-standard format min / max
        if max_t in {Decimal, date, datetime}:
            return (
                {"formatExclusiveMaximum": max_}
                if pred.exclusive_maximum
                else {"formatMaximum": max_}
            )
        else:
            return (
                {"exclusiveMaximum": max_}
                if pred.exclusive_maximum
                else {"maximum": max_}
            )
    elif isinstance(pred, EqualTo):
        # todo: is there a better way to do this than using enum?
        match_t = type(pred.match)
        if (
            match_t is str
            or match_t is int
            or match_t is None
            or match_t is float
            or match_t is bool
        ):
            choice = pred.match
        elif match_t is date:
            choice = pred.match.isoformat()
        elif match_t is datetime:
            choice = pred.match.isoformat()
        elif match_t is Decimal:
            choice = str(pred.match)
        elif match_t is UUID:
            choice = str(pred.match)
        elif match_t is bytes:
            choice = pred.match.decode("utf-8")
        else:
            raise TypeError(f"got unexpected type: {type(pred.match)}")

        return {"enum": [choice]}
    # objects
    elif isinstance(pred, MinKeys):
        return {"minProperties": pred.size}
    elif isinstance(pred, MaxKeys):
        return {"maxProperties": pred.size}
    # arrays
    elif isinstance(pred, MinItems):
        return {"minItems": pred.item_count}
    elif isinstance(pred, MaxItems):
        return {"maxItems": pred.item_count}
    elif isinstance(pred, UniqueItems):
        return {"uniqueItems": True}
    else:
        unhandled_type(pred)


def generate_schema_validator(
    to_schema_fn: ValidatorToSchema, obj: Validator[Any]
) -> Dict[str, Serializable]:
    r"""
    Produces a JSON Schema compatible Serializable from a ``Validator``

    :param to_schema_fn: the function to use for describing the Validator at the next
        depth -- for nested ``Validator``\s, ``Predicate``\s, and ``PredicateAsync``\s
    :param obj: the ``Validator`` being described
    :return: a JSON Schema compatible Serializable
    :raises AssertionError: only on a misconfiguration with ``OptionalValidator``\s
    :raises TypeError: if there's no implementation for a given type (this can usually be
        resolved with wrapper functions)
    """
    # caution, mutation below!
    if isinstance(obj, BoolValidator):
        return boolean_schema(to_schema_fn, obj)
    if isinstance(obj, StringValidator):
        return string_schema(to_schema_fn, obj)
    elif isinstance(obj, IntValidator):
        return integer_schema(to_schema_fn, obj)
    elif isinstance(obj, FloatValidator):
        return float_schema(to_schema_fn, obj)
    elif isinstance(obj, OptionalValidator):
        maybe_val_ret = to_schema_fn(obj.non_none_validator)
        if type(maybe_val_ret) is not dict:
            raise AssertionError("must be a dict")
        maybe_val_ret["nullable"] = True
        return maybe_val_ret
    elif isinstance(obj, UUIDValidator):
        return uuid_schema(to_schema_fn, obj)
    elif isinstance(obj, MapValidator):
        return map_of_schema(to_schema_fn, obj)
    elif isinstance(obj, RecordValidator):
        return obj_schema(to_schema_fn, obj)
    elif isinstance(obj, DataclassValidator):
        return dataclass_validator_schema(to_schema_fn, obj)
    elif isinstance(obj, TypedDictValidator):
        return typeddict_validator_schema(to_schema_fn, obj)
    elif isinstance(obj, NamedTupleValidator):
        return namedtuple_validator_schema(to_schema_fn, obj)
    elif isinstance(obj, EqualsValidator):
        return equals_schema(to_schema_fn, obj)
    elif isinstance(obj, KeyNotRequired):
        return to_schema_fn(obj.validator)
    elif isinstance(obj, DictValidatorAny):
        return dict_validator_schema(to_schema_fn, obj)
    elif isinstance(obj, ListValidator):
        return array_of_schema(to_schema_fn, obj)
    elif isinstance(obj, UniformTupleValidator):
        return array_of_schema(to_schema_fn, obj)
    elif isinstance(obj, IsDictValidator):
        return {"type": "object"}
    elif isinstance(obj, UnionValidator):
        return {"oneOf": [to_schema_fn(s) for s in obj.validators]}
    elif isinstance(obj, NTupleValidator):
        return {
            "description": f'a {len(obj.fields)}-tuple of the fields in "prefixItems"',
            "type": "array",
            "additionalItems": False,
            "maxItems": len(obj.fields),
            "minItems": len(obj.fields),
            "prefixItems": [to_schema_fn(s) for s in obj.fields],
        }
    elif isinstance(obj, DateValidator):
        return date_schema(to_schema_fn, obj)
    elif isinstance(obj, DatetimeValidator):
        return datetime_schema(to_schema_fn, obj)
    elif isinstance(obj, DecimalValidator):
        return decimal_schema(to_schema_fn, obj)
    elif isinstance(obj, BytesValidator):
        return bytes_schema(to_schema_fn, obj)
    elif isinstance(obj, CacheValidatorBase):
        return to_schema_fn(obj.validator)
    elif isinstance(obj, Lazy):
        raise TypeError(
            "`Lazy` type was not handled -- perhaps you need to use "
            "`generate_named_schema`?"
        )
    else:
        unhandled_type(obj)


def generate_schema_base(
    obj: Union[Validator[Any], Predicate[Any], PredicateAsync[Any]]
) -> Dict[str, Serializable]:
    if isinstance(obj, Validator):
        return generate_schema_validator(generate_schema_base, obj)
    elif isinstance(obj, (Predicate, PredicateAsync)):
        return generate_schema_predicate(obj)
    else:
        unhandled_type(obj)


def generate_named_schema_base(
    ref_location: str, schema_name: str, obj: AnyValidatorOrPredicate
) -> Dict[str, Serializable]:
    if isinstance(obj, Validator):
        to_schema_fn = partial(generate_named_schema_base, ref_location, schema_name)
        if isinstance(obj, Lazy):
            if obj.recurrent:
                return {"$ref": f"{ref_location}{schema_name}"}
            else:
                # cannot_proceed_from_here since the validator is a thunk
                return {}
        else:
            return generate_schema_validator(to_schema_fn, obj)
    elif isinstance(obj, (Predicate, PredicateAsync)):
        return generate_schema_predicate(obj)
    else:
        unhandled_type(obj)


def to_named_json_schema(
    schema_name: str,
    obj: AnyValidatorOrPredicate,
    ref_location: str = "#/components/schemas/",
) -> Dict[str, Serializable]:
    """
    Produces a ``Serializable`` from a ``Validator``, ``Predicate`` or ``PredicateAsync``
    that conforms to a valid JSON Schema. The root dict's only key is the schema name,
    which has the full JSON Schema as its value.

    :param schema_name: mainly required when recursive objects need to related to
        themselves
    :param obj: the Validator being described
    :param ref_location: where the schema name will be found if/when making a recursive
        reference to it
    :return: a ``Serializable`` compatible with JSON Schema
    """
    return {schema_name: generate_named_schema_base(ref_location, schema_name, obj)}


def to_json_schema(obj: AnyValidatorOrPredicate) -> Dict[str, Serializable]:
    """
    Produces a ``Serializable`` from a ``Validator``, ``Predicate`` or ``PredicateAsync``
    that conforms to a valid JSON Schema.

    :param obj: the Validator being described
    :return: a ``Serializable`` compatible with JSON Schema
    """
    return generate_schema_base(obj)



from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from typing import (
    Any,
    ClassVar,
    Hashable,
    List,
    Optional,
    Set,
    Sized,
    Tuple,
    Type,
    TypeVar,
)
from uuid import UUID

from koda import Thunk

from koda_validate._generics import A, Ret
from koda_validate._internal import _ResultTuple, _ToTupleValidator
from koda_validate.base import Predicate, Processor, Validator
from koda_validate.errors import PredicateErrs, TypeErr
from koda_validate.valid import Invalid, ValidationResult


class Lazy(Validator[Ret]):
    """
    Allows for specification of mutually recursive type definitions.
    """

    __match_args__ = (
        "validator",
        "recurrent",
    )

    def __init__(
        self,
        validator: Thunk[Validator[Ret]],
        recurrent: bool = True,
    ) -> None:
        """
        :param validator: the validator we actually want to use
        :param recurrent: whether this is being used in a recursive way. This
            is useful, so we can avoid infinite loops when traversing
            over validators (i.e. for openapi generation)
        """
        self.validator = validator
        self.recurrent = recurrent

    async def validate_async(self, data: Any) -> ValidationResult[Ret]:
        return await self.validator().validate_async(data)

    def __call__(self, data: Any) -> ValidationResult[Ret]:
        return self.validator()(data)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.validator == other.validator
            and self.recurrent == other.recurrent
        )

    def __repr__(self) -> str:
        return f"Lazy({repr(self.validator)}, recurrent={repr(self.recurrent)})"


ChoiceT = TypeVar("ChoiceT", bound=Hashable)


@dataclass
class Choices(Predicate[ChoiceT]):
    """
    A allow to check some ``Hashable`` type against a finite set of values.
    """

    choices: Set[ChoiceT]

    def __call__(self, val: ChoiceT) -> bool:
        return val in self.choices


MinMaxT = TypeVar("MinMaxT", int, float, Decimal, date, datetime)


@dataclass
class Min(Predicate[MinMaxT]):
    minimum: MinMaxT
    exclusive_minimum: bool = False

    def __call__(self, val: MinMaxT) -> bool:
        if self.exclusive_minimum:
            return val > self.minimum
        else:
            return val >= self.minimum


@dataclass
class Max(Predicate[MinMaxT]):
    maximum: MinMaxT
    exclusive_maximum: bool = False

    def __call__(self, val: MinMaxT) -> bool:
        if self.exclusive_maximum:
            return val < self.maximum
        else:
            return val <= self.maximum


Num = TypeVar("Num", int, float, Decimal)


@dataclass
class MultipleOf(Predicate[Num]):
    factor: Num

    def __call__(self, val: Num) -> bool:
        return val % self.factor == 0


# todo: expand types?
# note that we are allowing `float` because python allows float equivalence checks
# doesn't mean it's recommended to use it!
ExactMatchT = TypeVar(
    "ExactMatchT",
    bool,
    bytes,
    int,
    Decimal,
    str,
    float,
    date,
    datetime,
    UUID,
)


@dataclass
class EqualTo(Predicate[ExactMatchT]):
    match: ExactMatchT

    def __call__(self, val: ExactMatchT) -> bool:
        return val == self.match


@dataclass(init=False)
class EqualsValidator(_ToTupleValidator[ExactMatchT]):
    """
    Check if a value is of the same type as ``match`` *and* check that that
    value is equivalent.
    """

    match: ExactMatchT
    preprocessors: Optional[List[Processor[ExactMatchT]]] = None

    def __init__(
        self,
        match: ExactMatchT,
        preprocessors: Optional[List[Processor[ExactMatchT]]] = None,
    ) -> None:
        self.match = match
        self.preprocessors = preprocessors
        self.predicate: EqualTo[ExactMatchT] = EqualTo(match)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[ExactMatchT]:
        return self._validate_to_tuple(val)

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[ExactMatchT]:
        if (match_type := type(self.match)) == type(val):
            if self.preprocessors:
                for preprocess in self.preprocessors:
                    val = preprocess(val)

            if self.predicate(val):
                return True, val
            else:
                return False, Invalid(PredicateErrs([self.predicate]), val, self)
        else:
            return False, Invalid(TypeErr(match_type), val, self)


class AlwaysValid(_ToTupleValidator[A]):
    """
    Whatever value is submitted for validation will be returned as valid
    """

    __match_args__ = ()

    _instance: ClassVar[Optional["AlwaysValid[Any]"]] = None

    def __new__(cls) -> "AlwaysValid[A]":
        # make a singleton
        if cls._instance is None:
            cls._instance = super(AlwaysValid, cls).__new__(cls)
        return cls._instance

    def _validate_to_tuple(self, val: A) -> _ResultTuple[A]:
        return True, val

    async def _validate_to_tuple_async(self, val: A) -> _ResultTuple[A]:
        return True, val

    def __repr__(self) -> str:
        return "AlwaysValid()"


# Any must be the generic param here, but, AlwaysValid() can take on any generic type
always_valid: AlwaysValid[Any] = AlwaysValid()

ListOrTupleOrSetAny = TypeVar("ListOrTupleOrSetAny", List[Any], Tuple[Any, ...], Set[Any])


@dataclass
class MinItems(Predicate[ListOrTupleOrSetAny]):
    item_count: int

    def __call__(self, val: ListOrTupleOrSetAny) -> bool:
        return len(val) >= self.item_count


@dataclass
class MaxItems(Predicate[ListOrTupleOrSetAny]):
    item_count: int

    def __call__(self, val: ListOrTupleOrSetAny) -> bool:
        return len(val) <= self.item_count


@dataclass
class ExactItemCount(Predicate[Sized]):
    item_count: int

    def __call__(self, val: Sized) -> bool:
        return len(val) == self.item_count


@dataclass
class UniqueItems(Predicate[ListOrTupleOrSetAny]):
    """
    Works with both hashable and unhashable items.
    """

    def __call__(self, val: ListOrTupleOrSetAny) -> bool:
        hashable_items: Set[Tuple[Type[Any], Any]] = set()
        # slower lookups for unhashables
        unhashable_items: List[Tuple[Type[Any], Any]] = []
        for item in val:
            # needed to tell difference between things like
            # ints and bools
            typed_lookup = (type(item), item)
            try:
                if typed_lookup in hashable_items:
                    return False
                else:
                    hashable_items.add(typed_lookup)
            except TypeError:  # not hashable!
                if typed_lookup in unhashable_items:
                    return False
                else:
                    unhashable_items.append(typed_lookup)
        else:
            return True


# mypy has a problem with this for Tuple[Any, ...]
# for some types, you might need to use UniqueItems()
unique_items = UniqueItems()

StrOrBytes = TypeVar("StrOrBytes", str, bytes)


@dataclass
class MaxLength(Predicate[StrOrBytes]):
    length: int

    def __call__(self, val: StrOrBytes) -> bool:
        return len(val) <= self.length


@dataclass
class MinLength(Predicate[StrOrBytes]):
    length: int

    def __call__(self, val: StrOrBytes) -> bool:
        return len(val) >= self.length


@dataclass
class ExactLength(Predicate[StrOrBytes]):
    length: int

    def __call__(self, val: StrOrBytes) -> bool:
        return len(val) == self.length


@dataclass
class StartsWith(Predicate[StrOrBytes]):
    prefix: StrOrBytes

    def __call__(self, val: StrOrBytes) -> bool:
        return val.startswith(self.prefix)


@dataclass
class EndsWith(Predicate[StrOrBytes]):
    suffix: StrOrBytes

    def __call__(self, val: StrOrBytes) -> bool:
        return val.endswith(self.suffix)


@dataclass
class NotBlank(Predicate[StrOrBytes]):
    _instance: ClassVar[Optional["NotBlank[Any]"]] = None

    def __new__(cls) -> "NotBlank[StrOrBytes]":
        # make a singleton
        if cls._instance is None:
            cls._instance = super(NotBlank, cls).__new__(cls)
        return cls._instance

    def __call__(self, val: StrOrBytes) -> bool:
        return len(val.strip()) != 0


not_blank = NotBlank()


@dataclass
class Strip(Processor[StrOrBytes]):
    def __call__(self, val: StrOrBytes) -> StrOrBytes:
        return val.strip()


strip = Strip()


@dataclass
class UpperCase(Processor[StrOrBytes]):
    def __call__(self, val: StrOrBytes) -> StrOrBytes:
        return val.upper()


@dataclass
class LowerCase(Processor[StrOrBytes]):
    def __call__(self, val: StrOrBytes) -> StrOrBytes:
        return val.lower()


upper_case = UpperCase()

lower_case = LowerCase()



from typing import Dict, List, Tuple, Union

Serializable = Union[
    None,
    int,
    str,
    bool,
    float,
    List["Serializable"],
    Tuple["Serializable", ...],
    Dict[str, "Serializable"],
]



from dataclasses import dataclass
from typing import Any, Optional

from koda_validate._generics import A
from koda_validate._internal import (
    _ResultTuple,
    _ToTupleValidator,
    _union_validator,
    _union_validator_async,
)
from koda_validate.base import Validator
from koda_validate.coerce import Coercer
from koda_validate.errors import CoercionErr, TypeErr
from koda_validate.valid import Invalid


@dataclass
class NoneValidator(_ToTupleValidator[None]):
    coerce: Optional[Coercer[None]] = None

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[None]:
        if self.coerce:
            if self.coerce(val).is_just:
                return True, None
            else:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, type(None)), val, self
                )

        if val is None:
            return True, None
        else:
            return False, Invalid(TypeErr(type(None)), val, self)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[None]:
        return self._validate_to_tuple(val)


none_validator = NoneValidator()


class OptionalValidator(_ToTupleValidator[Optional[A]]):
    """
    We have a value for a key, but it can be null (None)
    """

    __match_args__ = ("non_none_validator", "none_validator")

    def __init__(
        self,
        validator: Validator[A],
        *,
        none_validator: Validator[None] = NoneValidator(),
    ) -> None:
        self.non_none_validator = validator
        self.none_validator = none_validator
        self.validators = (none_validator, validator)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[Optional[A]]:
        return await _union_validator_async(self, self.validators, val)

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[Optional[A]]:
        return _union_validator(self, self.validators, val)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and other.non_none_validator == self.non_none_validator
            and other.none_validator == self.none_validator
        )

    def __repr__(self) -> str:
        return f"OptionalValidator({repr(self.non_none_validator)})"



from typing import Any, Dict, List, Optional, Union

from koda_validate._generics import A
from koda_validate._internal import (
    _async_predicates_warning,
    _repr_helper,
    _ResultTuple,
    _ToTupleValidator,
    _wrap_async_validator,
    _wrap_sync_validator,
)
from koda_validate.base import Predicate, PredicateAsync, Validator
from koda_validate.coerce import Coercer
from koda_validate.errors import CoercionErr, IndexErrs, PredicateErrs, TypeErr
from koda_validate.valid import Invalid


class ListValidator(_ToTupleValidator[List[A]]):
    __match_args__ = ("item_validator", "predicates", "predicates_async", "coerce")

    def __init__(
        self,
        item_validator: Validator[A],
        *,
        predicates: Optional[List[Predicate[List[A]]]] = None,
        predicates_async: Optional[List[PredicateAsync[List[A]]]] = None,
        coerce: Optional[Coercer[List[Any]]] = None,
    ) -> None:
        self.item_validator = item_validator
        self.predicates = predicates
        self.predicates_async = predicates_async
        self._disallow_synchronous = bool(predicates_async)
        self.coerce = coerce

        self._wrapped_item_validator_sync = _wrap_sync_validator(item_validator)
        self._wrapped_item_validator_async = _wrap_async_validator(item_validator)

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[List[A]]:
        if self._disallow_synchronous:
            _async_predicates_warning(self.__class__)

        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, list), val, self
                )
            else:
                coerced_val: List[Any] = coerced.val

        elif type(val) is list:
            coerced_val = val
        else:
            return False, Invalid(TypeErr(list), val, self)

        if self.predicates:
            list_errors: List[Union[Predicate[List[A]], PredicateAsync[List[A]]]] = [
                pred for pred in self.predicates if not pred.__call__(coerced_val)
            ]

            if list_errors:
                return False, Invalid(PredicateErrs(list_errors), coerced_val, self)

        return_list: List[A] = []
        index_errs: Dict[int, Invalid] = {}
        for i, item in enumerate(coerced_val):
            is_valid, item_result = self._wrapped_item_validator_sync(item)

            if not is_valid:
                index_errs[i] = item_result  # type: ignore
            elif not index_errs:
                return_list.append(item_result)  # type: ignore

        if index_errs:
            return False, Invalid(IndexErrs(index_errs), coerced_val, self)
        else:
            return True, return_list

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[List[A]]:
        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, list), val, self
                )
            else:
                coerced_val: List[Any] = coerced.val

        elif type(val) is list:
            coerced_val = val
        else:
            return False, Invalid(TypeErr(list), val, self)

        predicate_errors: List[
            Union[Predicate[List[Any]], PredicateAsync[List[Any]]]
        ] = []
        if self.predicates:
            predicate_errors.extend(
                [pred for pred in self.predicates if not pred(coerced_val)]
            )

        if self.predicates_async is not None:
            for pred_async in self.predicates_async:
                if not await pred_async.validate_async(coerced_val):
                    predicate_errors.append(pred_async)

        if predicate_errors:
            return False, Invalid(PredicateErrs(predicate_errors), coerced_val, self)

        return_list: List[A] = []
        index_errs = {}
        for i, item in enumerate(coerced_val):
            (
                is_valid,
                item_result,
            ) = await self._wrapped_item_validator_async(item)

            if not is_valid:
                index_errs[i] = item_result
            elif not index_errs:
                return_list.append(item_result)  # type: ignore

        if index_errs:
            return False, Invalid(IndexErrs(index_errs), coerced_val, self)  # type: ignore  # noqa: E501
        else:
            return True, return_list

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) is type(self)
            and self.item_validator == other.item_validator
            and self.predicates == other.predicates
            and self.predicates_async == other.predicates_async
            and self.coerce == other.coerce
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [repr(self.item_validator)]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("predicates", self.predicates),
                    ("predicates_async", self.predicates_async),
                    ("coerce", self.coerce),
                ]
                if v
            ],
        )



from typing import Any

from koda import Just, Maybe, nothing

from koda_validate._generics import A
from koda_validate._internal import (
    _ResultTuple,
    _ToTupleValidator,
    _wrap_async_validator,
    _wrap_sync_validator,
)
from koda_validate.base import Validator
from koda_validate.errors import ContainerErr, TypeErr
from koda_validate.valid import Invalid


class MaybeValidator(_ToTupleValidator[Maybe[A]]):
    __match_args__ = ("validator",)

    def __init__(self, validator: Validator[A]):
        self.validator = validator
        self._validator_sync = _wrap_sync_validator(self.validator)
        self._validator_async = _wrap_async_validator(self.validator)

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[Maybe[A]]:
        if val is nothing:
            return True, nothing
        elif type(val) is Just:
            result = await self._validator_async(val.val)
            if result[0]:
                return True, Just(result[1])
            else:
                return False, Invalid(ContainerErr(result[1]), val, self)
        else:
            return False, Invalid(TypeErr(Maybe[Any]), val, self)  # type: ignore[misc]

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[Maybe[A]]:
        if val is nothing:
            return True, nothing
        elif type(val) is Just:
            result = self._validator_sync(val.val)
            if result[0]:
                return True, Just(result[1])
            else:
                return False, Invalid(ContainerErr(result[1]), val, self)
        else:
            return False, Invalid(TypeErr(Maybe[Any]), val, self)  # type: ignore[misc]

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.validator == other.validator

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.validator)})"



import inspect
import sys
from dataclasses import is_dataclass
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Dict,
    Optional,
    Protocol,
    Set,
    Type,
    TypeVar,
    cast,
    get_type_hints,
)

from koda import Just, Maybe, nothing

from koda_validate._internal import (
    _raise_cannot_define_validate_object_and_validate_object_async,
    _raise_validate_object_async_in_sync_mode,
    _repr_helper,
    _ResultTuple,
    _ToTupleValidator,
    _wrap_async_validator,
    _wrap_sync_validator,
)
from koda_validate.base import Validator
from koda_validate.coerce import Coercer
from koda_validate.errors import (
    CoercionErr,
    ErrType,
    ExtraKeysErr,
    KeyErrs,
    missing_key_err,
)
from koda_validate.typehints import get_typehint_validator
from koda_validate.valid import Invalid


class DataclassLike(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Any]]


_DCT = TypeVar("_DCT", bound=DataclassLike)


def dataclass_no_coerce(data_cls: Type[_DCT]) -> Coercer[Dict[Any, Any]]:
    def _fn(val: Any) -> Maybe[Dict[Any, Any]]:
        return Just(val.__dict__) if type(val) is data_cls else nothing

    return Coercer(_fn, {data_cls})


class DataclassValidator(_ToTupleValidator[_DCT]):
    """
    Takes a ``dataclass`` as an argument and derives a :class:`Validator`. Will validate
    against an instance of ``self.data_cls`` *or* a dictionary. Regardless of the input
    type, the result will be an instance of type ``self.data_cls``.

    Optional keys are determined by the presence of a default argument.

    Example:

    .. testcode:: dcexample

        from dataclasses import dataclass
        from typing import List
        from koda_validate import *

        @dataclass
        class Person:
            name: str
            hobbies: List[str]

        validator = DataclassValidator(Person)

    Usage:

    .. doctest:: dcexample

        >>> validator({"name": "Bob", "hobbies": ["eating", "coding", "sleeping"]})
        Valid(val=Person(name='Bob', hobbies=['eating', 'coding', 'sleeping']))


    :param data_cls: A ``dataclass`` class
    :param overrides: a dict whose keys define explicit validators
    :param validate_object: is run if all keys have been validated individually. If it
        returns ``None``, then there were no errors; otherwise it should return
        ``ErrType``
    :param validate_object_async: same as ``validate_object``, except that is runs
        asynchronously
    :param typehint_resolver: define this to override default inferred validators for
        types
    :param fail_on_unknown_keys: if True, this will fail if any keys not defined by
        ``self.data_cls`` are found. This will fail before any values are validated.
    :param coerce: a function that can control coercion
    :raises TypeError: should raise if non-``dataclass`` type is passed for ``data_cls``
    """

    __match_args__ = ("data_cls", "overrides", "fail_on_unknown_keys", "coerce")

    def __init__(
        self,
        data_cls: Type[_DCT],
        *,
        overrides: Optional[Dict[str, Validator[Any]]] = None,
        validate_object: Optional[Callable[[_DCT], Optional[ErrType]]] = None,
        validate_object_async: Optional[
            Callable[[_DCT], Awaitable[Optional[ErrType]]]
        ] = None,
        fail_on_unknown_keys: bool = False,
        typehint_resolver: Callable[[Any], Validator[Any]] = get_typehint_validator,
        coerce: Optional[Coercer[Dict[Any, Any]]] = None,
    ) -> None:
        if not is_dataclass(data_cls):
            raise TypeError("Must be a dataclass")
        self.data_cls = cast(Type[_DCT], data_cls)
        self.fail_on_unknown_keys = fail_on_unknown_keys
        self.overrides = overrides
        self.coerce = coerce
        if validate_object and validate_object_async:
            _raise_cannot_define_validate_object_and_validate_object_async()

        self.validate_object = validate_object
        self.validate_object_async = validate_object_async

        self._disallow_synchronous = bool(validate_object_async)

        keys_with_defaults: Set[str] = {
            k
            for k, v in inspect.signature(self.data_cls).parameters.items()
            if v.default != inspect.Parameter.empty
        }

        if sys.version_info >= (3, 9):
            type_hints = get_type_hints(self.data_cls, include_extras=True)
        else:
            type_hints = get_type_hints(self.data_cls)

        overrides = self.overrides or {}
        self.schema = {
            field: (
                overrides[field] if field in overrides else typehint_resolver(annotations)
            )
            for field, annotations in type_hints.items()
        }

        self.required_fields = []

        self._keys_set = set()
        self._fast_keys_sync = []
        self._fast_keys_async = []
        for key, val in self.schema.items():
            self._keys_set.add(key)
            is_required = key not in keys_with_defaults
            if is_required:
                self.required_fields.append(key)
            self._fast_keys_sync.append((key, _wrap_sync_validator(val), is_required))
            self._fast_keys_async.append((key, _wrap_async_validator(val), is_required))

        self._unknown_keys_err: ExtraKeysErr = ExtraKeysErr(set(self.schema.keys()))

    def _validate_to_tuple(self, val: Any) -> _ResultTuple[_DCT]:
        if self._disallow_synchronous:
            _raise_validate_object_async_in_sync_mode(self.__class__)

        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, dict), val, self
                )
            else:
                coerced_val: Dict[Any, Any] = coerced.val

        elif type(val) is dict:
            coerced_val = val
        elif type(val) is self.data_cls:
            coerced_val = val.__dict__
        else:
            return False, Invalid(
                CoercionErr(
                    {dict, self.data_cls},
                    self.data_cls,
                ),
                val,
                self,
            )

        if self.fail_on_unknown_keys:
            for key_ in coerced_val:
                if key_ not in self._keys_set:
                    return False, Invalid(self._unknown_keys_err, coerced_val, self)

        success_dict: Dict[Any, Any] = {}
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_sync:
            if key_ not in coerced_val:
                if key_required:
                    errs[key_] = Invalid(missing_key_err, coerced_val, self)
            else:
                success, new_val = validator(coerced_val[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    success_dict[key_] = new_val

        if errs:
            return False, Invalid(KeyErrs(errs), coerced_val, self)
        else:
            obj = self.data_cls(**success_dict)
            if self.validate_object and (result := self.validate_object(obj)):
                return False, Invalid(result, obj, self)

            return True, obj

    async def _validate_to_tuple_async(self, val: Any) -> _ResultTuple[_DCT]:
        if self.coerce:
            if not (coerced := self.coerce(val)).is_just:
                return False, Invalid(
                    CoercionErr(self.coerce.compatible_types, dict), val, self
                )
            else:
                coerced_val: Dict[Any, Any] = coerced.val

        elif type(val) is dict:
            coerced_val = val
        elif type(val) is self.data_cls:
            coerced_val = val.__dict__
        else:
            return False, Invalid(
                CoercionErr(
                    {dict, self.data_cls},
                    self.data_cls,
                ),
                val,
                self,
            )

        if self.fail_on_unknown_keys:
            # this seems to be faster than `for key_ in coerced_val.keys()`
            for key_ in coerced_val:
                if key_ not in self._keys_set:
                    return False, Invalid(self._unknown_keys_err, coerced_val, self)

        success_dict: Dict[Any, Any] = {}
        errs: Dict[Any, Invalid] = {}
        for key_, validator, key_required in self._fast_keys_async:
            if key_ not in coerced_val:
                if key_required:
                    errs[key_] = Invalid(missing_key_err, coerced_val, self)
            else:
                success, new_val = await validator(coerced_val[key_])

                if not success:
                    errs[key_] = new_val
                elif not errs:
                    success_dict[key_] = new_val

        if errs:
            return False, Invalid(KeyErrs(errs), coerced_val, self)
        else:
            obj = self.data_cls(**success_dict)
            if self.validate_object and (result := self.validate_object(obj)):
                return False, Invalid(result, obj, self)
            elif self.validate_object_async and (
                result_async := await self.validate_object_async(obj)
            ):
                return False, Invalid(result_async, obj, self)
            return True, obj

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) == type(other)
            and self.data_cls is other.data_cls
            and other.validate_object is self.validate_object
            and other.validate_object_async is self.validate_object_async
            and other.schema == self.schema
            and other.fail_on_unknown_keys == self.fail_on_unknown_keys
            and other.coerce == self.coerce
        )

    def __repr__(self) -> str:
        return _repr_helper(
            self.__class__,
            [repr(self.data_cls)]
            + [
                f"{k}={repr(v)}"
                for k, v in [
                    ("overrides", self.overrides),
                    ("validate_object", self.validate_object),
                    ("validate_object_async", self.validate_object_async),
                    ("coerce", self.coerce),
                    # note that this coincidentally works as we want:
                    # by default we don't fail on extra keys, so we don't
                    # show this in the repr if the default is defined
                    ("fail_on_unknown_keys", self.fail_on_unknown_keys),
                ]
                if v
            ],
        )



__all__ = (
    # serialization.py
    "Serializable",
    "SerializableErr",
    "to_serializable_errs",
    "to_json_schema",
    "to_named_json_schema",
)

from koda_validate.serialization.base import Serializable
from koda_validate.serialization.errors import SerializableErr, to_serializable_errs
from koda_validate.serialization.json_schema import to_json_schema, to_named_json_schema



from koda_validate._internal import _ToTupleStandardValidator


class BoolValidator(_ToTupleStandardValidator[bool]):
    r"""
    Validate a value is a ``bool``, and any extra refinement.

    If ``predicates_async`` is supplied, the ``__call__`` method should not be
    called -- only ``.validate_async`` should be used.

    :param predicates: any number of ``Predicate[bool]`` instances
    :param predicates_async: any number of ``PredicateAsync[bool]`` instances
    :param preprocessors: any number of ``Processor[bool]``, which will be run before
        :class:`Predicate`\s and :class:`PredicateAsync`\s are checked.
    :param coerce: a function that can control coercion
    """

    _TYPE = bool
